"""
.. module:: install

:Synopsis: Tools and script to install the external code and data packages needed by
           the Cobaya components to be used.
:Author: Jesus Torrado

"""
# Global
import os
import sys
import subprocess
import traceback
import logging
import shutil
from pkg_resources import parse_version

# Local
from cobaya.log import logger_setup, LoggedError
from cobaya.tools import create_banner, warn_deprecation, get_class, \
    write_packages_path_in_config_file
from cobaya.input import get_used_components, get_kind
from cobaya.conventions import _component_path, _code, _data, _external, _force, \
    _packages_path, _packages_path_arg, _packages_path_env, _yaml_extensions, \
    _install_skip_env, _packages_path_arg_posix
from cobaya.mpi import set_mpi_disabled
from cobaya.tools import resolve_packages_path

log = logging.getLogger(__name__.split(".")[-1])


# noinspection PyUnresolvedReferences
def install(*infos, **kwargs):
    if not log.root.handlers:
        logger_setup()
    path = kwargs.get("path")
    if not path:
        path = resolve_packages_path(infos)
    if not path:
        raise LoggedError(
            log, "No 'path' argument given, and none could be found in input infos "
                 "(as %r), the %r env variable or the config file. "
                 "Maybe specify one via a command line argument '-%s [...]'?",
            _packages_path, _packages_path_env, _packages_path_arg[0])
    abspath = os.path.abspath(path)
    log.info("Installing external packages at '%s'", abspath)
    kwargs_install = {"force": kwargs.get("force", False),
                      "no_progress_bars": kwargs.get("no_progress_bars")}
    for what in (_code, _data):
        kwargs_install[what] = kwargs.get(what, True)
        spath = os.path.join(abspath, what)
        if kwargs_install[what] and not os.path.exists(spath):
            try:
                os.makedirs(spath)
            except OSError:
                raise LoggedError(
                    log, "Could not create the desired installation folder '%s'", spath)
    failed_components = []
    skip_keywords = set(kwargs.get("skip", []) or [])
    skip_keywords_env = set(
        os.environ.get(_install_skip_env, "").replace(",", " ").lower().split())
    skip_keywords = skip_keywords.union(skip_keywords_env)
    for kind, components in get_used_components(*infos).items():
        for component in components:
            print()
            print(create_banner(kind + ":" + component, symbol="=", length=80))
            print()
            if _skip_helper(component.lower(), skip_keywords, skip_keywords_env, log):
                continue
            info = (next(info for info in infos if component in
                         info.get(kind, {}))[kind][component]) or {}
            if isinstance(info, str) or _external in info:
                log.warning("Component '%s' is a custom function. "
                            "Nothing to do.", component)
                continue
            try:
                imported_class = get_class(component, kind,
                                           component_path=info.pop(_component_path, None))
            except ImportError as e:
                log.error("Component '%s' not recognized. [%s]." % (component, e))
                failed_components += ["%s:%s" % (kind, component)]
                continue
            else:
                if _skip_helper(imported_class.__name__.lower(), skip_keywords,
                                skip_keywords_env, log):
                    continue
            is_installed = getattr(imported_class, "is_installed", None)
            if is_installed is None:
                log.info("%s.%s is a fully built-in component: nothing to do.",
                         kind, imported_class.__name__)
                continue
            if is_installed(path=abspath, **kwargs_install):
                log.info("External component already installed.")
                if kwargs.get("just_check", False):
                    continue
                if kwargs_install["force"]:
                    log.info("Forcing re-installation, as requested.")
                else:
                    log.info("Doing nothing.")
                    continue
            else:
                if kwargs.get("just_check", False):
                    log.info("NOT INSTALLED!")
                    continue
            try:
                install_this = getattr(imported_class, "install", None)
                success = install_this(path=abspath, **kwargs_install)
            except KeyboardInterrupt:
                raise
            except:
                traceback.print_exception(*sys.exc_info(), file=sys.stdout)
                log.error("An unknown error occurred. Delete the external packages "
                          "folder %r and try again. "
                          "Please, notify the developers if this error persists.",
                          abspath)
                success = False
            if success:
                log.info("Successfully installed!")
            else:
                log.error("Installation failed! Look at the error messages above. "
                          "Solve them and try again, or, if you are unable to solve, "
                          "install the packages required by this component manually.")
                failed_components += ["%s:%s" % (kind, component)]
                continue
            # test installation
            if not is_installed(path=abspath, **kwargs_install):
                log.error("Installation apparently worked, "
                          "but the subsequent installation test failed! "
                          "Look at the error messages above. "
                          "Solve them and try again, or, if you are unable to solve, "
                          "install the packages required by this component manually.")
                failed_components += ["%s:%s" % (kind, component)]
    if failed_components:
        bullet = "\n - "
        raise LoggedError(
            log, "The installation (or installation test) of some component(s) has "
                 "failed: %s\nCheck output of the installer of each component above "
                 "for precise error info.\n",
            bullet + bullet.join(failed_components))
    # Set the installation path in the global config file
    if not kwargs.get("no_set_global", False) and not kwargs.get("just_check", False):
        write_packages_path_in_config_file(abspath)
        log.info("The installation path has been written in the global config file.")


def _skip_helper(name, skip_keywords, skip_keywords_env, logger):
    try:
        this_skip_keyword = next(s for s in skip_keywords
                                 if s.lower() in name.lower())
        env_msg = (" in env var %r" % _install_skip_env
                   if this_skip_keyword in skip_keywords_env else "")
        logger.info("Skipping %r as per skip keyword %r" + env_msg,
                    name, this_skip_keyword)
        return True
    except StopIteration:
        return False


def download_file(filename, path, no_progress_bars=False, decompress=False, logger=None):
    logger = logger or logging.getLogger(__name__)
    try:
        from wget import download, bar_thermometer
        wget_kwargs = {"out": path, "bar":
            (bar_thermometer if not no_progress_bars else None)}
        filename = os.path.normpath(download(filename, **wget_kwargs))
        print("")
        logger.info('Downloaded filename %s' % filename)
    except Exception as e:
        logger.error(
            "Error downloading file '%s' to folder '%s': %s", filename, path, e)
        return False
    logger.debug('Got: %s' % filename)
    if not decompress:
        return True
    extension = os.path.splitext(filename)[-1][1:]
    try:
        if extension == "zip":
            from zipfile import ZipFile
            with ZipFile(filename, 'r') as zipObj:
                zipObj.extractall(path)
        else:
            import tarfile
            if extension == "tgz":
                extension = "gz"
            with tarfile.open(filename, "r:" + extension) as tar:
                tar.extractall(path)
        logger.debug('Decompressed: %s' % filename)
        os.remove(filename)
        return True
    except Exception as e:
        logger.error("Error decompressing downloaded file! Corrupt file? [%s]" % e)
        return False


def download_github_release(directory, repo_name, release_name, repo_rename=None,
                            no_progress_bars=False, logger=None):
    logger = logger or logging.getLogger(__name__)
    if "/" in repo_name:
        github_user = repo_name[:repo_name.find("/")]
        repo_name = repo_name[repo_name.find("/") + 1:]
    else:
        github_user = "CobayaSampler"
    if not os.path.exists(directory):
        os.makedirs(directory)
    filename = (r"https://github.com/" + github_user + "/" + repo_name +
                "/archive/" + release_name + ".tar.gz")
    if not download_file(filename, directory, decompress=True,
                         no_progress_bars=no_progress_bars, logger=logger):
        return False
    # Remove version number from directory name
    w_version = next(d for d in os.listdir(directory)
                     if (d.startswith(repo_name) and len(d) != len(repo_name)))
    repo_rename = repo_rename or repo_name
    repo_path = os.path.join(directory, repo_rename)
    if os.path.exists(repo_path):
        shutil.rmtree(repo_path)
    os.rename(os.path.join(directory, w_version), repo_path)
    logger.info("%s %s downloaded and decompressed correctly.", repo_name, release_name)
    return True


def pip_install(packages, upgrade=False):
    """
    Takes package name or list of them.

    Uses ``--user`` flag if does not appear to have write permission to python path

    Returns exit status.
    """
    if hasattr(packages, "split"):
        packages = [packages]
    cmd = [sys.executable, '-m', 'pip', 'install']
    if not os.access(os.path.dirname(sys.executable), os.W_OK):
        cmd += ['--user']
    if upgrade:
        cmd += ['--upgrade']
    res = subprocess.call(cmd + packages)
    if res:
        log.error("pip: error installing packages '%s'", packages)
    return res


def check_gcc_version(min_version="6.4", error_returns=None):
    try:
        version = subprocess.check_output(
            "gcc -dumpversion", shell=True, stderr=subprocess.STDOUT).decode().strip()
    except:
        return error_returns
    # Change in gcc >= 7: -dumpversion only dumps major version
    if "." not in version:
        version = subprocess.check_output(
            "gcc -dumpfullversion", shell=True, stderr=subprocess.STDOUT).decode().strip()
    return parse_version(str(min_version)) <= parse_version(version)


# Command-line script ####################################################################

def install_script():
    set_mpi_disabled(True)
    warn_deprecation()
    # Parse arguments
    import argparse
    parser = argparse.ArgumentParser(
        description="Cobaya's installation tool for external packages.")
    parser.add_argument("files_or_components", action="store", nargs="+",
                        metavar="input_file.yaml|component_name",
                        help="One or more input files or component names "
                             "(or simply 'cosmo' to install all the requisites for basic"
                             " cosmological runs)")
    parser.add_argument("-" + _packages_path_arg[0], "--" + _packages_path_arg_posix,
                        action="store", nargs=1, required=False,
                        metavar="/packages/path", default=[None],
                        help="Desired path where to install external packages. "
                             "Optional if one has been set globally or as an env variable"
                             " (run with '--show_%s' to check)." %
                             _packages_path_arg_posix)
    # MARKED FOR DEPRECATION IN v3.0
    modules = "modules"
    parser.add_argument("-" + modules[0], "--" + modules,
                        action="store", nargs=1, required=False,
                        metavar="/packages/path", default=[None],
                        help="To be deprecated! "
                             "Alias for %s, which should be used instead." %
                             _packages_path_arg_posix)
    # END OF DEPRECATION BLOCK -- CONTINUES BELOW!
    output_show_packages_path = resolve_packages_path()
    if output_show_packages_path and os.environ.get(_packages_path_env):
        output_show_packages_path += " (from env variable %r)" % _packages_path_env
    elif output_show_packages_path:
        output_show_packages_path += " (from config file)"
    else:
        output_show_packages_path = "(Not currently set.)"
    parser.add_argument("--show-" + _packages_path_arg_posix, action="version",
                        version=output_show_packages_path,
                        help="Prints default external packages installation folder "
                             "and exits.")
    parser.add_argument("-" + _force[0], "--" + _force, action="store_true",
                        default=False,
                        help="Force re-installation of apparently installed packages.")
    parser.add_argument("--skip", action="store", nargs="*",
                        metavar="keyword",
                        help="Keywords of components that will be skipped during "
                             "installation.")
    parser.add_argument("--no-progress-bars", action="store_true", default=False,
                        help="No progress bars shown. Shorter logs (used in Travis).")
    parser.add_argument("--just-check", action="store_true", default=False,
                        help="Just check whether components are installed.")
    parser.add_argument("--no-set-global", action="store_true", default=False,
                        help="Do not store the installation path for later runs.")
    group_just = parser.add_mutually_exclusive_group(required=False)
    group_just.add_argument("-C", "--just-code", action="store_false", default=True,
                            help="Install code of the components.", dest=_data)
    group_just.add_argument("-D", "--just-data", action="store_false", default=True,
                            help="Install data of the components.", dest=_code)
    arguments = parser.parse_args()
    # Configure the logger ASAP
    logger_setup()
    logger = logging.getLogger(__name__.split(".")[-1])
    # Gather requests
    infos = []
    for f in arguments.files_or_components:
        if f.lower() == "cosmo":
            logger.info("Installing basic cosmological packages.")
            from cobaya.cosmo_input import install_basic
            infos += [install_basic]
        elif f.lower() == "cosmo-tests":
            logger.info("Installing *tested* cosmological packages.")
            from cobaya.cosmo_input import install_tests
            infos += [install_tests]
        elif os.path.splitext(f)[1].lower() in _yaml_extensions:
            from cobaya.input import load_input
            infos += [load_input(f)]
        else:
            try:
                kind = get_kind(f)
                infos += [{kind: {f: None}}]
            except Exception:
                logger.warning("Could not identify component %r. Skipping.", f)
    if not infos:
        logger.info("Nothing to install.")
        return
    # MARKED FOR DEPRECATION IN v3.0
    if getattr(arguments, modules) != [None]:
        logger.warning("*DEPRECATION*: -m/--modules will be deprecated in favor of "
                       "-%s/--%s in the next version. Please, use that one instead.",
                       _packages_path_arg[0], _packages_path_arg_posix)
        # BEHAVIOUR TO BE REPLACED BY ERROR:
        if getattr(arguments, _packages_path_arg) == [None]:
            setattr(arguments, _packages_path_arg, getattr(arguments, modules))
    # END OF DEPRECATION BLOCK
    # Launch installer
    install(*infos, path=getattr(arguments, _packages_path_arg)[0],
            **{arg: getattr(arguments, arg)
               for arg in ["force", _code, _data, "no_progress_bars", "just_check",
                           "no_set_global", "skip"]})


if __name__ == '__main__':
    install_script()



# Notes about writing more presets:
# ---------------------------------
# - all parameter names below are PLANCK parameter names. They are substituted by the
#   theory-code-specific ones in `create_input`
# - don't use extra_args for precision parameters! because if the same precision param
#   is mentioned twice at the same time in different fields with different values, there
#   is no facility to take the max (or min). Instead, codify precision needs in terms of
#   requirements in the .must_provide method of the cosmo code.

# Global
from copy import deepcopy

# Local
from cobaya.conventions import kinds, partag as p, _params

_camb = "camb"
_classy = "classy"
_desc = "desc"
_comment = "note"
_extra_args = "extra_args"
_error_msg = "error_msg"
_none = "(None)"

# Theory codes
theory = {_camb: None, _classy: None}

# Primordial perturbations
primordial = dict([
    ("SFSR", {
        _desc: "Adiabatic scalar perturbations, power law spectrum",
        kinds.theory: {_camb: None, _classy: None},
        _params: dict([
            ["logA", dict([
                (p.prior, dict([("min", 1.61), ("max", 3.91)])),
                (p.ref, dict([(p.dist, "norm"), ("loc", 3.05), ("scale", 0.001)])),
                (p.proposal, 0.001), (p.latex, r"\log(10^{10} A_\mathrm{s})"),
                (p.drop, True)])],
            ("As", dict([
                (p.value, "lambda logA: 1e-10*np.exp(logA)"),
                (p.latex, r"A_\mathrm{s}")])),
            ("ns", dict([
                (p.prior, dict([("min", 0.8), ("max", 1.2)])),
                (p.ref, dict([(p.dist, "norm"), ("loc", 0.965), ("scale", 0.004)])),
                (p.proposal, 0.002), (p.latex, r"n_\mathrm{s}")]))])})])
primordial.update(dict([
    ["SFSR_DESpriors", {
        _desc: "Adiabatic scalar perturbations, power law + running spectrum "
               "-- DESpriors",
        kinds.theory: {_camb: None, _classy: None},
        _params: dict([
            ["As_1e9", dict([
                [p.prior, dict([["min", 0.5], ["max", 5]])],
                [p.ref,
                 dict([[p.dist, "norm"], ["loc", 2.1], ["scale", 0.5]])],
                [p.proposal, 0.25], [p.latex, r"10^9 A_\mathrm{s})"],
                [p.drop, True], [p.renames, "A"]])],
            ["As", dict([
                [p.value, "lambda As_1e9: 1e-9 * As_1e9"],
                [p.latex, r"A_\mathrm{s}"]])],
            ["ns", primordial["SFSR"][_params]["ns"]]])}]]))
primordial.update(dict([
    ["SFSR_run", {
        _desc: "Adiabatic scalar perturbations, power law + running spectrum",
        kinds.theory: {_camb: None, _classy: None},
        _params: dict(
            (list(primordial["SFSR"][_params].items()) +
             [["nrun", dict([
                 [p.prior, dict([["min", -1], ["max", 1]])],
                 [p.ref,
                  dict([[p.dist, "norm"], ["loc", 0], ["scale", 0.005]])],
                 [p.proposal, 0.001], [p.latex, r"n_\mathrm{run}"]])]]))}]]))
primordial.update(dict([
    ["SFSR_runrun", {
        _desc: "Adiabatic scalar perturbations, power law + 2nd-order running spectrum",
        kinds.theory: {_camb: None, _classy: None},
        _params: dict(
            (list(primordial["SFSR_run"][_params].items()) +
             [["nrunrun", dict([
                 [p.prior, dict([["min", -1], ["max", 1]])],
                 [p.ref,
                  dict([[p.dist, "norm"], ["loc", 0], ["scale", 0.002]])],
                 [p.proposal, 0.001],
                 [p.latex, r"n_\mathrm{run,run}"]])]]))}]]))
primordial.update(dict([
    ["SFSR_t", {
        _desc: "Adiabatic scalar+tensor perturbations, "
               "power law spectrum (inflation consistency)",
        kinds.theory: {_camb: {_extra_args: {"nt": None}},
                       _classy: {_extra_args: {"n_t": "scc", "alpha_t": "scc"}}},
        _params: dict(
            (list(primordial["SFSR"][_params].items()) +
             [["r", dict([
                 [p.prior, dict([["min", 0], ["max", 3]])],
                 [p.ref,
                  dict([[p.dist, "norm"], ["loc", 0], ["scale", 0.03]])],
                 [p.proposal, 0.03], [p.latex, r"r_{0.05}"]])]]))}]]))
primordial.update(dict([
    ["SFSR_t_nrun", {
        _desc: "Adiabatic scalar+tensor perturbations, "
               "power law + running spectrum (inflation consistency)",
        kinds.theory: {_camb:
                           {_extra_args: primordial["SFSR_t"][kinds.theory][_camb][
                               _extra_args]},
                       _classy:
                           {_extra_args: primordial["SFSR_t"][kinds.theory][_classy][
                               _extra_args]}},
        _params: dict(
            (list(primordial["SFSR_run"][_params].items()) +
             list(primordial["SFSR_t"][_params].items())))}]]))

# Geometry
geometry = dict([
    ["flat", {
        _desc: "Flat FLRW universe",
        kinds.theory: {_camb: None, _classy: None}}],
    ["omegak", {
        _desc: "FLRW model with varying curvature (prior on Omega_k)",
        kinds.theory: {_camb: None, _classy: None},
        _params: dict([
            ["omegak", dict([
                [p.prior, dict([["min", -0.3], ["max", 0.3]])],
                [p.ref,
                 dict([[p.dist, "norm"], ["loc", -0.009], ["scale", 0.001]])],
                [p.proposal, 0.001], [p.latex, r"\Omega_k"]])]])}], ])

# Hubble parameter constraints
H0_min, H0_max = 20, 100
hubble = dict([
    ["H", {
        _desc: "Hubble parameter",
        kinds.theory: {_camb: None, _classy: None},
        _params: dict([
            ["H0", dict([
                [p.prior, dict([["min", H0_min], ["max", H0_max]])],
                [p.ref, dict([[p.dist, "norm"], ["loc", 67], ["scale", 2]])],
                [p.proposal, 2], [p.latex, r"H_0"]])]])}],
    ["H_DESpriors", {
        _desc: "Hubble parameter (reduced range for DES and lensing-only constraints)",
        kinds.theory: {_camb: None, _classy: None},
        _params: dict([
            ["H0", dict([
                [p.prior, dict([["min", 55], ["max", 91]])],
                [p.ref, dict([[p.dist, "norm"], ["loc", 67], ["scale", 2]])],
                [p.proposal, 2], [p.latex, r"H_0"]])]])}],
    ["sound_horizon_last_scattering", {
        _desc: "Angular size of the sound horizon at last scattering "
               "(approximate, if using CAMB)",
        kinds.theory: {
            _camb: {
                _params: dict([
                    ["theta_MC_100", dict([
                        [p.prior, dict([["min", 0.5], ["max", 10]])],
                        [p.ref,
                         dict([[p.dist, "norm"], ["loc", 1.04109],
                               ["scale", 0.0004]])],
                        [p.proposal, 0.0002],
                        [p.latex, r"100\theta_\mathrm{MC}"],
                        [p.drop, True], [p.renames, "theta"]])],
                    ["cosmomc_theta", dict([
                        [p.value, "lambda theta_MC_100: 1.e-2*theta_MC_100"],
                        [p.derived, False]])],
                    ["H0", {p.latex: r"H_0", "min": H0_min, "max": H0_max}]]),
                _extra_args: dict([["theta_H0_range", [H0_min, H0_max]]])},
            _classy: {
                _params: dict([
                    ["theta_s_1e2", dict([
                        [p.prior, dict([["min", 0.5], ["max", 10]])],
                        [p.ref, dict([
                            [p.dist, "norm"], ["loc", 1.0416], ["scale", 0.0004]])],
                        [p.proposal, 0.0002],
                        [p.latex, r"100\theta_\mathrm{s}"],
                        [p.drop, True]])],
                    ["100*theta_s", dict([
                        [p.value, "lambda theta_s_1e2: theta_s_1e2"],
                        [p.derived, False]])],
                    ["H0", {p.latex: r"H_0"}]])}}}]])

# Matter sector (minus light species)
N_eff_std = 3.046
nu_mass_fac = 94.0708
matter = dict([
    ["omegab_h2, omegac_h2", {
        _desc: "Flat prior on Omega*h^2 for baryons and cold dark matter",
        kinds.theory: {_camb: None, _classy: None},
        _params: dict([
            ["omegabh2", dict([
                [p.prior, dict([["min", 0.005], ["max", 0.1]])],
                [p.ref,
                 dict([[p.dist, "norm"], ["loc", 0.0224], ["scale", 0.0001]])],
                [p.proposal, 0.0001], [p.latex, r"\Omega_\mathrm{b} h^2"]])],
            ["omegach2", dict([
                [p.prior, dict([["min", 0.001], ["max", 0.99]])],
                [p.ref,
                 dict([[p.dist, "norm"], ["loc", 0.120], ["scale", 0.001]])],
                [p.proposal, 0.0005], [p.latex, r"\Omega_\mathrm{c} h^2"]])],
            ["omegam", {p.latex: r"\Omega_\mathrm{m}"}]])}],
    ["Omegab, Omegam", {
        _desc: "Flat prior on Omega for baryons and total matter",
        kinds.theory: {_camb: None, _classy: None},
        _params: dict([
            ["omegab", dict([
                [p.prior, dict([["min", 0.03], ["max", 0.07]])],
                [p.ref,
                 dict([[p.dist, "norm"], ["loc", 0.0495], ["scale", 0.004]])],
                [p.proposal, 0.004], [p.latex, r"\Omega_\mathrm{b}"],
                [p.drop, True]])],
            ["omegam", dict([
                [p.prior, dict([["min", 0.1], ["max", 0.9]])],
                [p.ref,
                 dict([[p.dist, "norm"], ["loc", 0.316], ["scale", 0.02]])],
                [p.proposal, 0.02], [p.latex, r"\Omega_\mathrm{m}"],
                [p.drop, True]])],
            ["omegabh2", dict([
                [p.value, "lambda omegab, H0: omegab*(H0/100)**2"],
                [p.latex, r"\Omega_\mathrm{b} h^2"]])],
            ["omegach2", dict([
                [p.value, ("lambda omegam, omegab, mnu, H0: "
                           "(omegam-omegab)*(H0/100)**2-(mnu*(%g/3)**0.75)/%g" % (
                               N_eff_std, nu_mass_fac))],
                [p.latex, r"\Omega_\mathrm{c} h^2"]])]
        ])}]])
for m in matter.values():
    m[_params]["omegamh2"] = dict([
        [p.derived, "lambda omegam, H0: omegam*(H0/100)**2"],
        [p.latex, r"\Omega_\mathrm{m} h^2"]])

# Neutrinos and other extra matter
neutrinos = dict([
    ["one_heavy_planck", {
        _desc: "Two massless nu and one with m=0.06. Neff=3.046",
        kinds.theory: {
            _camb: {_extra_args: {"num_massive_neutrinos": 1, "nnu": 3.046},
                    _params: dict([["mnu", 0.06]])},
            _classy: {_extra_args: {"N_ncdm": 1, "N_ur": 2.0328},
                      _params: dict([
                          ["m_ncdm", {p.value: 0.06, p.renames: "mnu"}]])}}}],
    ["varying_mnu", {
        _desc: "Varying total mass of 3 degenerate nu's, with N_eff=3.046",
        kinds.theory: {
            _camb: {_extra_args: {"num_massive_neutrinos": 3, "nnu": 3.046},
                    _params: dict([
                        ["mnu", dict([
                            [p.prior, dict([["min", 0], ["max", 5]])],
                            [p.ref, dict([
                                [p.dist, "norm"], ["loc", 0.02], ["scale", 0.1]])],
                            [p.proposal, 0.03], [p.latex, r"\sum m_\nu"]])]])},
            _classy: {_extra_args: {"N_ncdm": 1, "deg_ncdm": 3, "N_ur": 0.00641},
                      _params: dict([
                          ["m_ncdm", dict([
                              [p.prior, dict([["min", 0], ["max", 1.667]])],
                              [p.ref, dict([
                                  [p.dist, "norm"], ["loc", 0.0067],
                                  ["scale", 0.033]])],
                              [p.proposal, 0.01], [p.latex, r"m_\nu"]])],
                          ["mnu", dict([[p.derived, "lambda m_ncdm: 3 * m_ncdm"],
                                        [p.latex, r"\sum m_\nu"]])]])}}}],
    ["varying_Neff", {
        _desc: "Varying Neff with two massless nu and one with m=0.06",
        kinds.theory: {
            _camb: {_extra_args: {"num_massive_neutrinos": 1},
                    _params: dict([
                        ["mnu", 0.06],
                        ["nnu", dict([
                            [p.prior, dict([["min", 0.05], ["max", 10]])],
                            [p.ref, dict([
                                [p.dist, "norm"], ["loc", 3.046], ["scale", 0.05]])],
                            [p.proposal, 0.05],
                            [p.latex, r"N_\mathrm{eff}"]])]])},
            _classy: {_extra_args: {"N_ncdm": 1},
                      _params: dict([
                          ["m_ncdm",
                           dict([[p.value, 0.06], [p.renames, "mnu"]])],
                          ["N_ur", dict([
                              [p.prior, dict([["min", 0.0001], ["max", 9]])],
                              [p.ref, dict([
                                  [p.dist, "norm"], ["loc", 2.0328],
                                  ["scale", 0.05]])],
                              [p.proposal, 0.05],
                              [p.latex, r"N_\mathrm{ur}"]])],
                          ["nnu", dict([
                              [p.derived, "lambda Neff: Neff"],
                              [p.latex, r"N_\mathrm{eff}"]])]])}}}]])
neutrinos.update(dict([
    ["varying_mnu_Neff", {
        _desc: "Varying Neff and total mass of 3 degenerate nu's",
        kinds.theory: {
            _camb: {
                _extra_args: {"num_massive_neutrinos": 3},
                _params: dict([
                    ["mnu", deepcopy(
                        neutrinos["varying_mnu"][kinds.theory]["camb"][_params]["mnu"])],
                    ["nnu", deepcopy(
                        neutrinos["varying_Neff"][kinds.theory]["camb"][_params][
                            "nnu"])]])},
            #            _classy: {
            #                _extra_args: {"N_ncdm": 1, "deg_ncdm": 3},
            #                _params: dict([
            #                    ["m_ncdm", deepcopy(
            #                        neutrinos["varying_mnu"][_theory]["classy"][_params]["m_ncdm"])],
            #                    ["mnu", deepcopy(
            #                        neutrinos["varying_mnu"][_theory]["classy"][_params]["mnu"])],
            #                    ["N_ur", deepcopy(
            #                        neutrinos["varying_Neff"][_theory]["classy"][_params]["N_ur"])],
            #                    ["nnu", deepcopy(
            #                        neutrinos["varying_Neff"][_theory]["classy"][_params]["nnu"])]
            #                ])}
        }}]]))
# neutrinos["varying_mnu_Neff"][_theory][_classy][_params]["N_ur"][p.ref]["loc"] = 0.00641

#    # ["varying_Neff+1sterile", {
#    #     _desc: "Varying Neff plus 1 sterile neutrino (SM nu's with m=0,0,0.06)",
#    #     _theory: {
#    #         _camb: {_extra_args: dict(
#    #             list(neutrinos["varying_Neff"][_theory][_camb][_extra_args].items()) +
#    #             [["accuracy_level", 1.2]])}},
#    #     _params: dict([
#    #         ["nnu", dict([
#    #             [p.prior, dict([["min", 3.046], ["max", 10]])],
#    #             [p.ref, dict([[p.dist, "norm"], ["loc", 3.046], ["scale", 0.05]])],
#    #             [p.proposal, 0.05], [p.latex, r"N_\mathrm{eff}"]])],
#    #         ["meffsterile", dict([
#    #             [p.prior, dict([["min", 0], ["max", 3]])#,
#    #             [p.ref, dict([[p.dist, "norm"], ["loc", 0.1], ["scale", 0.1]])],
#    #             [p.proposal, 0.03],
#    #             [p.latex, r"m_{\nu,\mathrm{sterile}}^\mathrm{eff}"]])]])}]

# Dark Energy
dark_energy = dict([
    ["lambda", {
        _desc: "Cosmological constant (w=-1)",
        kinds.theory: {_camb: None, _classy: None},
        _params: dict([
            ["omegal", {p.latex: r"\Omega_\Lambda"}]])}],
    ["de_w", {
        _desc: "Varying constant eq of state",
        kinds.theory: {_camb: None,
                       _classy: {_params: {"Omega_Lambda": 0}}},
        _params: dict([
            ["w", dict([
                [p.prior, dict([["min", -3], ["max", -0.333]])],
                [p.ref,
                 dict([[p.dist, "norm"], ["loc", -0.99], ["scale", 0.02]])],
                [p.proposal, 0.02], [p.latex, r"w_\mathrm{DE}"]])]])}],
    ["de_w_wa", {
        _desc: "Varying constant eq of state with w(a) = w0 + (1-a) wa",
        kinds.theory: {_camb: None,
                       _classy: {_params: {"Omega_Lambda": 0}}},
        _params: dict([
            ["w", dict([
                [p.prior, dict([["min", -3], ["max", 1]])],
                [p.ref,
                 dict([[p.dist, "norm"], ["loc", -0.99], ["scale", 0.02]])],
                [p.proposal, 0.02], [p.latex, r"w_{0,\mathrm{DE}}"]])],
            ["wa", dict([
                [p.prior, dict([["min", -3], ["max", 2]])],
                [p.ref, dict([[p.dist, "norm"], ["loc", 0], ["scale", 0.05]])],
                [p.proposal, 0.05], [p.latex, r"w_{a,\mathrm{DE}}"]])]])}]])

# BBN
bbn_derived_camb = dict([
    ["YpBBN", dict([[p.latex, r"Y_P^\mathrm{BBN}"]])],
    ["DHBBN", dict([[p.derived, r"lambda DH: 10**5*DH"],
                    [p.latex, r"10^5 \mathrm{D}/\mathrm{H}"]])]])
bbn = dict([
    ["consistency", {
        _desc: "Primordial Helium fraction inferred from BBN consistency",
        kinds.theory: {_camb: {_params: bbn_derived_camb},
                       _classy: None},
        _params: dict([
            ["yheused", {p.latex: r"Y_\mathrm{P}"}]])}],
    ["YHe_des_y1", {
        _desc: "Fixed Y_P = 0.245341 (used in DES Y1)",
        kinds.theory: {_camb: None,
                       _classy: None},
        _params: dict([
            ["yhe", 0.245341]])}],
    ["YHe", {
        _desc: "Varying primordial Helium fraction",
        kinds.theory: {_camb: None,
                       _classy: None},
        _params: dict([
            ["yhe", dict([
                [p.prior, dict([["min", 0.1], ["max", 0.5]])],
                [p.ref,
                 dict([[p.dist, "norm"], ["loc", 0.237], ["scale", 0.006]])],
                [p.proposal, 0.006], [p.latex, r"Y_\mathrm{P}"]])]])}], ])

# Reionization
reionization = dict([
    ["std", {
        _desc: "Standard reio, lasting delta_z=0.5",
        kinds.theory: {_camb: None, _classy: None},
        _params: dict([
            ["tau", dict([
                [p.prior, dict([["min", 0.01], ["max", 0.8]])],
                [p.ref,
                 dict([[p.dist, "norm"], ["loc", 0.055], ["scale", 0.006]])],
                [p.proposal, 0.003], [p.latex, r"\tau_\mathrm{reio}"]])],
            ["zrei", {p.latex: r"z_\mathrm{re}"}]])}],
    ["gauss_prior", {
        _desc: "Standard reio, lasting delta_z=0.5, gaussian prior around tau=0.07",
        kinds.theory: {_camb: None, _classy: None},
        _params: dict([
            ["tau", dict([
                [p.prior,
                 dict([[p.dist, "norm"], ["loc", 0.07], ["scale", 0.02]])],
                [p.ref,
                 dict([[p.dist, "norm"], ["loc", 0.07], ["scale", 0.01]])],
                [p.proposal, 0.005], [p.latex, r"\tau_\mathrm{reio}"]])],
            ["zrei", {p.latex: r"z_\mathrm{re}"}]])}],
    ["irrelevant", {
        _desc: "Irrelevant (NB: only valid for non-CMB or CMB-marged datasets!)",
        kinds.theory: {_camb: None, _classy: None},
        _params: dict([])}], ])

# EXPERIMENTS ############################################################################
base_precision = {_camb: {"halofit_version": "mead"},
                  _classy: {"non linear": "hmcode", "hmcode_min_k_max": 20}}
cmb_precision = deepcopy(base_precision)
cmb_precision[_camb].update({"bbn_predictor": "PArthENoPE_880.2_standard.dat",
                             "lens_potential_accuracy": 1})
cmb_sampler_recommended = {"mcmc": {
    "drag": True, "oversample_power": 0.4, "proposal_scale": 1.9}}

like_cmb = dict([
    [_none, {}],
    ["planck_2018", {
        _desc: "Planck 2018 (Polarized CMB + lensing)",
        _comment: None,
        kinds.sampler: cmb_sampler_recommended,
        kinds.theory: {theo: {_extra_args: cmb_precision[theo]}
                       for theo in [_camb, _classy]},
        kinds.likelihood: dict([
            ["planck_2018_lowl.TT", None],
            ["planck_2018_lowl.EE", None],
            ["planck_2018_highl_plik.TTTEEE", None],
            ["planck_2018_lensing.clik", None]])}],
    ["planck_2018_bk15", {
        _desc: "Planck 2018 (Polarized CMB + lensing) + Bicep/Keck-Array 2015",
        kinds.sampler: cmb_sampler_recommended,
        kinds.theory: {theo: {_extra_args: cmb_precision[theo]}
                       for theo in [_camb, _classy]},
        kinds.likelihood: dict([
            ["planck_2018_lowl.TT", None],
            ["planck_2018_lowl.EE", None],
            ["planck_2018_highl_plik.TTTEEE", None],
            ["planck_2018_lensing.clik", None],
            ["bicep_keck_2015", None]])}],
    ["planck_2018_CMBmarged_lensing", {
        _desc: "Planck 2018 CMB-marginalized lensing",
        kinds.sampler: cmb_sampler_recommended,
        kinds.theory: {theo: {_extra_args: cmb_precision[theo]}
                       for theo in [_camb, _classy]},
        kinds.likelihood: dict([
            ["planck_2018_lensing.CMBMarged", None]])}],
])
like_cmb["planck_2018_bk15"][_comment] = like_cmb["planck_2018"][_comment]
# Add common CMB derived parameters
for name, m in like_cmb.items():
    # Don't add the derived parameter to the no-CMB case!
    if not m:
        continue
    if _params not in m:
        m[_params] = dict()
    m[_params].update(dict([
        ["sigma8", {p.latex: r"\sigma_8"}],
        ["s8h5", dict([
            [p.derived, "lambda sigma8, H0: sigma8*(H0*1e-2)**(-0.5)"],
            [p.latex, r"\sigma_8/h^{0.5}"]])],
        ["s8omegamp5", dict([
            [p.derived, "lambda sigma8, omegam: sigma8*omegam**0.5"],
            [p.latex, r"\sigma_8 \Omega_\mathrm{m}^{0.5}"]])],
        ["s8omegamp25", dict([
            [p.derived, "lambda sigma8, omegam: sigma8*omegam**0.25"],
            [p.latex, r"\sigma_8 \Omega_\mathrm{m}^{0.25}"]])],
        ["A", dict([
            [p.derived, "lambda As: 1e9*As"],
            [p.latex, r"10^9 A_\mathrm{s}"]])],
        ["clamp", dict([
            [p.derived, "lambda As, tau: 1e9*As*np.exp(-2*tau)"],
            [p.latex, r"10^9 A_\mathrm{s} e^{-2\tau}"]])],
        ["age", dict([
            [p.latex, r"{\rm{Age}}/\mathrm{Gyr}"]])],
        ["rdrag", dict([
            [p.latex, r"r_\mathrm{drag}"]])]]))
    if "cmbmarged" in name.lower():
        m[_params].pop("A")
        m[_params].pop("clamp")
# Some more, in case we want to add them at some point, described in
# https://wiki.cosmos.esa.int/planckpla2015/images/b/b9/Parameter_tag_definitions_2015.pdf
#    "zstar":       {"latex": r"z_*"},
#    "rstar":       {"latex": r"r_*"},
#    "thetastar":   {"latex": r"100\theta_*"},
#    "DAstar":      {"latex": r"D_\mathrm{A}/\mathrm{Gpc}"},
#    "zdrag":       {"latex": r"z_\mathrm{drag}"},
#    "kd":          {"latex": r"k_\mathrm{D}"},
#    "thetad":      {"latex": r"100\theta_\mathrm{D}"},
#    "zeq":         {"latex": r"z_\mathrm{eq}"},
#    "keq":         {"latex": r"k_\mathrm{eq}"},
#    "thetaeq":     {"latex": r"100\theta_\mathrm{eq}"},
#    "thetarseq":   {"latex": r"100\theta_\mathrm{s,eq}"},

like_bao = dict([
    [_none, {}],
    ["BAO_planck_2018", {
        _desc: "Baryon acoustic oscillation data from DR12, MGS and 6DF",
        kinds.theory: {_camb: None, _classy: None},
        kinds.likelihood: dict([
            ["bao.sixdf_2011_bao", None],
            ["bao.sdss_dr7_mgs", None],
            ["bao.sdss_dr12_consensus_bao", None]])}],
])

like_des = dict([
    [_none, {}],
    ["des_y1_clustering", {
        _desc: "Galaxy clustering from DES Y1",
        kinds.theory: {theo: {_extra_args: base_precision[theo]}
                       for theo in [_camb, _classy]},
        kinds.likelihood: dict([
            ["des_y1.clustering", None]])}],
    ["des_y1_galaxy_galaxy", {
        _desc: "Galaxy-galaxy lensing from DES Y1",
        kinds.theory: {theo: {_extra_args: base_precision[theo]}
                       for theo in [_camb, _classy]},
        kinds.likelihood: dict([
            ["des_y1.galaxy_galaxy", None]])}],
    ["des_y1_shear", {
        _desc: "Cosmic shear data from DES Y1",
        kinds.theory: {theo: {_extra_args: base_precision[theo]}
                       for theo in [_camb, _classy]},
        kinds.likelihood: dict([
            ["des_y1.shear", None]])}],
    ["des_y1_joint", {
        _desc: "Combination of galaxy clustering and weak lensing data from DES Y1",
        kinds.theory: {theo: {_extra_args: base_precision[theo]}
                       for theo in [_camb, _classy]},
        kinds.likelihood: dict([
            ["des_y1.joint", None]])}],
])

like_sn = dict([
    [_none, {}],
    ["Pantheon", {
        _desc: "Supernovae data from the Pantheon sample",
        kinds.theory: {_camb: None, _classy: None},
        kinds.likelihood: dict([
            ["sn.pantheon", None]])}],
])

like_H0 = dict([
    [_none, {}],
    ["Riess2018a", {
        _desc: "Local H0 measurement from Riess et al. 2018a (used in Planck 2018)",
        kinds.theory: {_camb: None, _classy: None},
        kinds.likelihood: dict([
            ["H0.riess2018a", None]])}],
    ["Riess201903", {
        _desc: "Local H0 measurement from Riess et al. 2019",
        kinds.theory: {_camb: None, _classy: None},
        kinds.likelihood: dict([
            ["H0.riess201903", None]])}],
])

# SAMPLERS ###############################################################################

sampler = dict([
    ["MCMC", {
        _desc: "MCMC sampler with covmat learning",
        kinds.sampler: {"mcmc": {"covmat": "auto"}}}],
    ["PolyChord", {
        _desc: "Nested sampler, affine invariant and multi-modal",
        kinds.sampler: {"polychord": None}}], ])

# PRESETS ################################################################################

planck_base_model = {
    "primordial": "SFSR",
    "geometry": "flat",
    "hubble": "sound_horizon_last_scattering",
    "matter": "omegab_h2, omegac_h2",
    "neutrinos": "one_heavy_planck",
    "dark_energy": "lambda",
    "bbn": "consistency",
    "reionization": "std"}
default_sampler = {"sampler": "MCMC"}

preset = dict([
    [_none, {_desc: "(No preset chosen)"}],
    # Pure CMB #######################################################
    ["planck_2018_camb", {
        _desc: "Planck 2018 with CAMB",
        "theory": _camb,
        "like_cmb": "planck_2018"}],
    ["planck_2018_classy", {
        _desc: "Planck 2018 with CLASS",
        "theory": _classy,
        "like_cmb": "planck_2018"}],
    ["planck_2018_bicep_camb", {
        _desc: "Planck 2018 + BK15 (with tensor modes) with CAMB",
        "theory": _camb,
        "primordial": "SFSR_t",
        "like_cmb": "planck_2018_bk15"}],
    ["planck_2018_bicep_classy", {
        _desc: "Planck 2018 + BK15 (with tensor modes) with CLASS",
        "theory": _classy,
        "primordial": "SFSR_t",
        "like_cmb": "planck_2018_bk15"}],
    # CMB+BAO ######################################################
    ["planck_2018_BAO_camb", {
        _desc: "Planck 2018 + BAO with CAMB",
        "theory": _camb,
        "like_cmb": "planck_2018",
        "like_bao": "BAO_planck_2018"}],
    ["planck_2018_BAO_classy", {
        _desc: "Planck 2018 + BAO with CLASS",
        "theory": _classy,
        "like_cmb": "planck_2018",
        "like_bao": "BAO_planck_2018"}],
    # CMB+BAO+SN ###################################################
    ["planck_2018_BAO_SN_camb", {
        _desc: "Planck 2018 + BAO + SN with CAMB",
        "theory": _camb,
        "like_cmb": "planck_2018",
        "like_bao": "BAO_planck_2018",
        "like_sn": "Pantheon"}],
    ["planck_2018_BAO_SN_classy", {
        _desc: "Planck 2018 + BAO + SN with CLASS",
        "theory": _classy,
        "like_cmb": "planck_2018",
        "like_bao": "BAO_planck_2018",
        "like_sn": "Pantheon"}],
    # CMB+DES+BAO+SN ###################################################
    ["planck_2018_DES_BAO_SN_camb", {
        _desc: "Planck 2018 + DESjoint + BAO + SN with CAMB",
        "theory": _camb,
        "like_cmb": "planck_2018",
        "like_bao": "BAO_planck_2018",
        "like_des": "des_y1_joint",
        "like_sn": "Pantheon"}],
    ["planck_2018_DES_BAO_SN_classy", {
        _desc: "Planck 2018 + DESjoint + BAO + SN with CLASS",
        "theory": _classy,
        "like_cmb": "planck_2018",
        "like_bao": "BAO_planck_2018",
        "like_des": "des_y1_joint",
        "like_sn": "Pantheon"}],
])

# Add planck baseline model
for pre in preset.values():
    pre.update(
        {field: value for field, value in planck_base_model.items() if field not in pre})
    pre.update(default_sampler)

# Lensing-only ###################################################
preset.update(dict([
    [_none, {_desc: "(No preset chosen)"}],
    ["planck_2018_DES_lensingonly_camb", {
        _desc: "Planck 2018 + DES Y1 lensing-only with CAMB",
        "theory": _camb,
        "like_cmb": "planck_2018_CMBmarged_lensing",
        "like_des": "des_y1_shear"}],
    ["planck_2018_DES_lensingonly_classy", {
        _desc: "Planck 2018 + DES Y1 lensing-only with CLASS",
        "theory": _classy,
        "like_cmb": "planck_2018_CMBmarged_lensing",
        "like_des": "des_y1_shear"}],
]))

lensingonly_model = {
    "primordial": "SFSR_DESpriors",
    "geometry": "flat",
    "hubble": "H_DESpriors",
    "matter": "Omegab, Omegam",
    "neutrinos": "one_heavy_planck",
    "dark_energy": "lambda",
    "bbn": "YHe_des_y1",
    "reionization": "irrelevant"}

# Add planck baseline model
for name, pre in preset.items():
    if "lensingonly" in name:
        pre.update(
            {field: value for field, value in lensingonly_model.items() if
             field not in pre})
        pre.update(default_sampler)

# BASIC INSTALLATION #####################################################################
install_basic = {
    kinds.theory: {_camb: None, _classy: None},
    kinds.likelihood: {
        "planck_2018_lowl.TT": None,
        "planck_2018_lensing.native": None,
        "bicep_keck_2015": None,
        "sn.pantheon": None,
        "bao.sdss_dr12_consensus_final": None,
        "des_y1.joint": None}}

install_tests = deepcopy(install_basic)
install_tests[kinds.likelihood].update({"planck_2015_lowl": None,
                                        "planck_2018_highl_plik.TT_unbinned": None,
                                        "planck_2018_highl_plik.TT_lite_native": None,
                                        "planck_2018_highl_CamSpec.TT": None,
                                        "planck_2018_highl_CamSpec.TT_native": None,
                                        })



"""
.. module:: run

:Synopsis: Main scripts to run the sampling
:Author: Jesus Torrado

"""

# Global
from typing import Mapping
import logging

# Local
from cobaya import __version__
from cobaya.conventions import kinds, _prior, _params, _packages_path, _output_prefix, \
    _debug, _debug_file, _resume, _timing, _debug_default, _force, _post, _test_run, \
    _yaml_extensions, _packages_path_arg, \
    _packages_path_arg_posix
from cobaya.output import get_output, split_prefix, get_info_path
from cobaya.model import Model
from cobaya.sampler import get_sampler_name_and_class, check_sampler_info
from cobaya.log import logger_setup, LoggedError
from cobaya.yaml import yaml_dump
from cobaya.input import update_info
from cobaya.mpi import import_MPI, is_main_process, set_mpi_disabled
from cobaya.tools import warn_deprecation, recursive_update, sort_cosmetic, \
    check_deprecated_modules_path
from cobaya.post import post


def run(info):
    # This function reproduces the model-->output-->sampler pipeline one would follow
    # when instantiating by hand, but alters the order to performs checks and dump info
    # as early as possible, e.g. to check if resuming possible or `force` needed.
    assert isinstance(info, Mapping), (
        "The first argument must be a dictionary with the info needed for the run. "
        "If you were trying to pass the name of an input file instead, "
        "load it first with 'cobaya.input.load_input', "
        "or, if you were passing a yaml string, load it with 'cobaya.yaml.yaml_load'.")
    logger_setup(info.get(_debug), info.get(_debug_file))
    logger_run = logging.getLogger(__name__.split(".")[-1])
    # MARKED FOR DEPRECATION IN v3.0
    # BEHAVIOUR TO BE REPLACED BY ERROR:
    check_deprecated_modules_path(info)
    # END OF DEPRECATION BLOCK
    # 1. Prepare output driver, if requested by defining an output_prefix
    output = get_output(output_prefix=info.get(_output_prefix),
                        resume=info.get(_resume), force=info.get(_force))
    # 2. Update the input info with the defaults for each component
    updated_info = update_info(info)
    if logging.root.getEffectiveLevel() <= logging.DEBUG:
        # Dump only if not doing output (otherwise, the user can check the .updated file)
        if not output and is_main_process():
            logger_run.info(
                "Input info updated with defaults (dumped to YAML):\n%s",
                yaml_dump(sort_cosmetic(updated_info)))
    # 3. If output requested, check compatibility if existing one, and dump.
    # 3.1 First: model only
    output.check_and_dump_info(info, updated_info, cache_old=True,
                               ignore_blocks=[kinds.sampler])
    # 3.2 Then sampler -- 1st get the last sampler mentioned in the updated.yaml
    # TODO: ideally, using Minimizer would *append* to the sampler block.
    #       Some code already in place, but not possible at the moment.
    try:
        last_sampler = list(updated_info[kinds.sampler])[-1]
        last_sampler_info = {last_sampler: updated_info[kinds.sampler][last_sampler]}
    except (KeyError, TypeError):
        raise LoggedError(logger_run, "No sampler requested.")
    sampler_name, sampler_class = get_sampler_name_and_class(last_sampler_info)
    check_sampler_info(
        (output.reload_updated_info(use_cache=True) or {}).get(kinds.sampler),
        updated_info[kinds.sampler], is_resuming=output.is_resuming())
    # Dump again, now including sampler info
    output.check_and_dump_info(info, updated_info, check_compatible=False)
    # Check if resumable run
    sampler_class.check_force_resume(
        output, info=updated_info[kinds.sampler][sampler_name])
    # 4. Initialize the posterior and the sampler
    with Model(updated_info[_params], updated_info[kinds.likelihood],
               updated_info.get(_prior), updated_info.get(kinds.theory),
               packages_path=info.get(_packages_path), timing=updated_info.get(_timing),
               allow_renames=False, stop_at_error=info.get("stop_at_error", False)) \
            as model:
        # Re-dump the updated info, now containing parameter routes and version info
        updated_info = recursive_update(updated_info, model.info())
        output.check_and_dump_info(None, updated_info, check_compatible=False)
        sampler = sampler_class(updated_info[kinds.sampler][sampler_name],
                                model, output, packages_path=info.get(_packages_path))
        # Re-dump updated info, now also containing updates from the sampler
        updated_info[kinds.sampler][sampler.get_name()] = \
            recursive_update(
                updated_info[kinds.sampler][sampler.get_name()], sampler.info())
        # TODO -- maybe also re-dump model info, now possibly with measured speeds
        # (waiting until the camb.transfers issue is solved)
        output.check_and_dump_info(None, updated_info, check_compatible=False)
        if info.get(_test_run, False):
            logger_run.info("Test initialization successful! "
                            "You can probably run now without `--%s`.", _test_run)
            return updated_info, sampler
        # Run the sampler
        sampler.run()
    return updated_info, sampler


# Command-line script
def run_script():
    warn_deprecation()
    import os
    import argparse
    parser = argparse.ArgumentParser(description="Cobaya's run script.")
    parser.add_argument("input_file", nargs=1, action="store", metavar="input_file.yaml",
                        help="An input file to run.")
    parser.add_argument("-" + _packages_path_arg[0], "--" + _packages_path_arg_posix,
                        action="store", nargs=1, metavar="/packages/path", default=[None],
                        help="Path where external packages were installed.")
    # MARKED FOR DEPRECATION IN v3.0
    modules = "modules"
    parser.add_argument("-" + modules[0], "--" + modules,
                        action="store", nargs=1, required=False,
                        metavar="/packages/path", default=[None],
                        help="To be deprecated! "
                             "Alias for %s, which should be used instead." %
                             _packages_path_arg_posix)
    # END OF DEPRECATION BLOCK -- CONTINUES BELOW!
    parser.add_argument("-" + _output_prefix[0], "--" + _output_prefix,
                        action="store", nargs=1, metavar="/some/path", default=[None],
                        help="Path and prefix for the text output.")
    parser.add_argument("-" + _debug[0], "--" + _debug, action="store_true",
                        help="Produce verbose debug output.")
    continuation = parser.add_mutually_exclusive_group(required=False)
    continuation.add_argument("-" + _resume[0], "--" + _resume, action="store_true",
                              help="Resume an existing chain if it has similar info "
                                   "(fails otherwise).")
    continuation.add_argument("-" + _force[0], "--" + _force, action="store_true",
                              help="Overwrites previous output, if it exists "
                                   "(use with care!)")
    parser.add_argument("--%s" % _test_run, action="store_true",
                        help="Initialize model and sampler, and exit.")
    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument("--no-mpi", action='store_true',
                        help="disable MPI when mpi4py installed but MPI does "
                             "not actually work")
    arguments = parser.parse_args()
    if arguments.no_mpi or getattr(arguments, _test_run, False):
        set_mpi_disabled()
    if any((os.path.splitext(f)[0] in ("input", "updated"))
           for f in arguments.input_file):
        raise ValueError("'input' and 'updated' are reserved file names. "
                         "Please, use a different one.")
    load_input = import_MPI(".input", "load_input")
    given_input = arguments.input_file[0]
    if any(given_input.lower().endswith(ext) for ext in _yaml_extensions):
        info = load_input(given_input)
        output_prefix_cmd = getattr(arguments, _output_prefix)[0]
        output_prefix_input = info.get(_output_prefix)
        info[_output_prefix] = output_prefix_cmd or output_prefix_input
    else:
        # Passed an existing output_prefix? Try to find the corresponding *.updated.yaml
        updated_file = get_info_path(*split_prefix(given_input), kind="updated")
        try:
            info = load_input(updated_file)
        except IOError:
            raise ValueError("Not a valid input file, or non-existent run to resume")
        # We need to update the output_prefix to resume the run *where it is*
        info[_output_prefix] = given_input
        # If input given this way, we obviously want to resume!
        info[_resume] = True
    # solve packages installation path cmd > env > input
    # MARKED FOR DEPRECATION IN v3.0
    if getattr(arguments, modules) != [None]:
        logger_setup()
        logger = logging.getLogger(__name__.split(".")[-1])
        logger.warning("*DEPRECATION*: -m/--modules will be deprecated in favor of "
                       "-%s/--%s in the next version. Please, use that one instead.",
                       _packages_path_arg[0], _packages_path_arg_posix)
        # BEHAVIOUR TO BE REPLACED BY ERROR:
        if getattr(arguments, _packages_path_arg) == [None]:
            setattr(arguments, _packages_path_arg, getattr(arguments, modules))
    # BEHAVIOUR TO BE REPLACED BY ERROR:
    check_deprecated_modules_path(info)
    # END OF DEPRECATION BLOCK
    info[_packages_path] = \
        getattr(arguments, _packages_path_arg)[0] or info.get(_packages_path)
    info[_debug] = getattr(arguments, _debug) or info.get(_debug, _debug_default)
    info[_test_run] = getattr(arguments, _test_run, False)
    # If any of resume|force given as cmd args, ignore those in the input file
    resume_arg, force_arg = [getattr(arguments, arg) for arg in [_resume, _force]]
    if any([resume_arg, force_arg]):
        info[_resume], info[_force] = resume_arg, force_arg
    if _post in info:
        post(info)
    else:
        run(info)



"""
.. module:: BoltzmannBase

:Synopsis: Template for Cosmological theory codes.
           Mostly here to document how to compute and get observables.
:Author: Jesus Torrado

"""
import numpy as np
from scipy.interpolate import RectBivariateSpline
from typing import Mapping, Iterable

# Local
from cobaya.theory import Theory
from cobaya.tools import deepcopy_where_possible
from cobaya.log import LoggedError
from cobaya.conventions import _c_km_s, empty_dict

H_units_conv_factor = {"1/Mpc": 1, "km/s/Mpc": _c_km_s}


class BoltzmannBase(Theory):
    _get_z_dependent: callable  # defined by inheriting classes
    renames: Mapping[str, str] = empty_dict

    def initialize(self):

        # Dict of named tuples to collect requirements and computation methods
        self.collectors = {}
        # Additional input parameters to pass to CAMB, and attributes to set_ manually
        self.extra_args = deepcopy_where_possible(self.extra_args) or {}
        self._must_provide = None

    def initialize_with_params(self):
        self.check_no_repeated_input_extra()

    def get_allow_agnostic(self):
        return True

    def translate_param(self, p):
        return self.renames.get(p, p)

    def get_param(self, p):
        translated = self.translate_param(p)
        for pool in ["params", "derived", "derived_extra"]:
            value = (self._current_state[pool] or {}).get(translated, None)
            if value is not None:
                return value

        raise LoggedError(self.log, "Parameter not known: '%s'", p)

    def _norm_vars_pairs(self, vars_pairs, name):
        # Empty list: default to *total matter*: CMB + Baryon + MassiveNu
        vars_pairs = vars_pairs or [2 * ["delta_tot"]]
        if not isinstance(vars_pairs, Iterable):
            raise LoggedError(self.log, "vars_pairs must be an iterable of pairs "
                                        "of variable names: got '%r' for %s",
                              vars_pairs, name)
        if isinstance(list(vars_pairs)[0], str):
            vars_pairs = [vars_pairs]
        pairs = set()
        for pair in vars_pairs:
            if len(pair) != 2 or not all(isinstance(x, str) for x in pair):
                raise LoggedError(self.log,
                                  "Cannot understand vars_pairs '%r' for %s",
                                  vars_pairs, name)
            pairs.add(tuple(sorted(pair)))
        return pairs

    def _check_args(self, req, value, vals):
        for par in vals:
            if par not in value:
                raise LoggedError(self.log, "%s must specify %s in requirements",
                                  req, par)

    def must_provide(self, **requirements):
        r"""
        Specifies the quantities that this Boltzmann code is requested to compute.

        Typical requisites in Cosmology (as keywords, case insensitive):

        - ``Cl={...}``: CMB lensed power spectra, as a dictionary ``{spectrum:l_max}``,
          where the possible spectra are combinations of "t", "e", "b" and "p"
          (lensing potential). Get with :func:`~BoltzmannBase.get_Cl`.
        - **[BETA: CAMB only; notation may change!]** ``source_Cl={...}``:
          :math:`C_\ell` of given sources with given windows, e.g.:
          ``source_name: {"function": "spline"|"gaussian", [source_args]``;
          for now, ``[source_args]`` follow the notation of ``CAMBSources``.
          If can also take ``lmax: [int]``, ``limber: True`` if Limber approximation
          desired, and ``non_linear: True`` if non-linear contributions requested.
          Get with :func:`~BoltzmannBase.get_source_Cl`.
        - ``Pk_interpolator={...}``: Matter power spectrum interpolator in :math:`(z, k)`.
          Takes ``"z": [list_of_evaluated_redshifts]``, ``"k_max": [k_max]``,
          ``"extrap_kmax": [max_k_max_extrapolated]``, ``"nonlinear": [True|False]``,
          ``"vars_pairs": [["delta_tot", "delta_tot"], ["Weyl", "Weyl"], [...]]}``.
          Non-linear contributions are included by default. Note that the nonlinear setting
          determines whether nonlinear corrections are calculated; the get_Pk_interpolator
          function also has a nonlinear argument to specify if you want the linear or
          nonlinear spectrum returned (to have both linear and non-linear spectra
          available request a tuple (False,True) for the nonlinear argument).
          All ``k`` values should be in units of ``1/Mpc``.
        - ``Pk_grid={...}``: similar to Pk_interpolator except that rather than returning
          a bicubic spline object it returns the raw power spectrum grid as a (k, z, PK)
          set of arrays.
        - ``sigma_R{...}``: RMS linear fluctuation in spheres of radius R at redshifts z.
          Takes ``"z": [list_of_evaluated_redshifts]``, ``"k_max": [k_max]``,
          ``"vars_pairs": [["delta_tot", "delta_tot"],  [...]]}``,
          ``"R": [list_of_evaluated_R]``. Note that R is in Mpc, not h^{-1} Mpc.
        - ``Hubble={'z': [z_1, ...]}``: Hubble rate at the requested redshifts.
          Get it with :func:`~BoltzmannBase.get_Hubble`.
        - ``angular_diameter_distance={'z': [z_1, ...]}``: Physical angular
          diameter distance to the redshifts requested. Get it with
          :func:`~BoltzmannBase.get_angular_diameter_distance`.
        - ``comoving_radial_distance={'z': [z_1, ...]}``: Comoving radial distance
          from us to the redshifts requested. Get it with
          :func:`~BoltzmannBase.get_comoving_radial_distance`.
        - ``fsigma8={'z': [z_1, ...]}``: Structure growth rate
          :math:`f\sigma_8` at the redshifts requested. Get it with
          :func:`~BoltzmannBase.get_fsigma8`.
        - ``k_max=[...]``: Fixes the maximum comoving wavenumber considered.
        - **Other derived parameters** that are not included in the input but whose
          value the likelihood may need.

        """
        super().must_provide(**requirements)
        self._must_provide = self._must_provide or dict.fromkeys(self.output_params)
        # Accumulate the requirements across several calls in a safe way;
        # e.g. take maximum of all values of a requested precision parameter
        for k, v in requirements.items():
            # Products and other computations
            if k == "Cl":
                current = self._must_provide.get(k, {})
                self._must_provide[k] = {cl: max(current.get(cl, 0), v.get(cl, 0))
                                         for cl in set(current).union(v)}
            elif k == 'sigma_R':
                self._check_args(k, v, ('z', 'R'))
                for pair in self._norm_vars_pairs(v.pop("vars_pairs", []), k):
                    k = ("sigma_R",) + pair
                    current = self._must_provide.get(k, {})
                    self._must_provide[k] = {
                        "R": np.sort(np.unique(np.concatenate(
                            (current.get("R", []), np.atleast_1d(v["R"]))))),
                        "z": np.unique(np.concatenate(
                            (current.get("z", []), np.atleast_1d(v["z"])))),
                        "k_max": max(current.get("k_max", 0),
                                     v.get("k_max", 2 / np.min(v["R"])))}
            elif k in ("Pk_interpolator", "Pk_grid"):
                # arguments are all identical, collect all in Pk_grid
                self._check_args(k, v, ('z', 'k_max'))
                redshifts = v.pop("z")
                k_max = v.pop("k_max")
                nonlin = v.pop("nonlinear", True)
                if not isinstance(nonlin, Iterable):
                    nonlin = [nonlin]
                for var_pair in self._norm_vars_pairs(v.pop("vars_pairs", []), k):
                    for nonlinear in nonlin:
                        k = ("Pk_grid", bool(nonlinear)) + var_pair
                        current = self._must_provide.get(k, {})
                        self._must_provide[k] = dict(
                            nonlinear=nonlinear,
                            z=np.unique(np.concatenate((current.get("z", []),
                                                        np.atleast_1d(redshifts)))),
                            k_max=max(current.get("k_max", 0), k_max), **v)
            elif k == "source_Cl":
                if k not in self._must_provide:
                    self._must_provide[k] = {}
                if "sources" not in v:
                    raise LoggedError(
                        self.log, "Needs a 'sources' key, containing a dict with every "
                                  "source name and definition")
                # Check that no two sources with equal name but diff specification
                # for source, window in v["sources"].items():
                #     if source in (getattr(self, "sources", {}) or {}):
                #         # TODO: improve this test!!!
                #         # (e.g. 2 z-vectors that fulfill np.allclose would fail a == test)
                #         if window != self.sources[source]:
                #             raise LoggedError(
                #                 self.log,
                #                 "Source %r requested twice with different specification: "
                #                 "%r vs %r.", window, self.sources[source])
                self._must_provide[k].update(v)
            elif k in ["Hubble", "angular_diameter_distance",
                       "comoving_radial_distance", "fsigma8"]:
                if k not in self._must_provide:
                    self._must_provide[k] = {}
                self._must_provide[k]["z"] = np.unique(np.concatenate(
                    (self._must_provide[k].get("z", []), v["z"])))
            # Extra derived parameters and other unknown stuff (keep capitalization)
            elif v is None:
                self._must_provide[k] = None
            else:
                raise LoggedError(self.log, "Unknown required product: '%s'.", k)

    def requested(self):
        """
        Returns the full set of requested cosmological products and parameters.
        """
        return self._must_provide

    def check_no_repeated_input_extra(self):
        """
        Checks that there are no repeated parameters between input and extra.

        Should be called at initialisation, and at the end of every call to must_provide()
        """
        common = set(self.input_params).intersection(set(self.extra_args))
        if common:
            raise LoggedError(
                self.log, "The following parameters appear both as input parameters and "
                          "as extra arguments: %s. Please, remove one of the definitions "
                          "of each.", common)

    def get_Cl(self, ell_factor=False, units="muK2"):
        r"""
        Returns a dictionary of lensed CMB power spectra and the lensing potential ``pp``
        power spectrum.

        Set the units with the keyword ``units='1'|'muK2'|'K2'`` (default: 'muK2',
        except for the lensing potential power spectrum, which is always unitless).

        If ``ell_factor=True`` (default: False), multiplies the spectra by
        :math:`\ell(\ell+1)/(2\pi)` (or by :math:`\ell^2(\ell+1)^2/(2\pi)` in the case of
        the lensing potential ``pp`` spectrum).
        """
        pass

    def get_Hubble(self, z, units="km/s/Mpc"):
        r"""
        Returns the Hubble rate at the given redshifts.

        The redshifts must be a subset of those requested when
        :func:`~BoltzmannBase.must_provide` was called.

        The available units are ``km/s/Mpc`` (i.e. ``c*H(Mpc^-1)``) and ``1/Mpc``.
        """
        try:
            return self._get_z_dependent("Hubble", z) * H_units_conv_factor[units]
        except KeyError:
            raise LoggedError(
                self.log, "Units not known for H: '%s'. Try instead one of %r.",
                units, list(H_units_conv_factor))

    def get_angular_diameter_distance(self, z):
        r"""
        Returns the physical angular diameter distance to the given redshifts in Mpc.

        The redshifts must be a subset of those requested when
        :func:`~BoltzmannBase.must_provide` was called.
        """
        return self._get_z_dependent("angular_diameter_distance", z)

    def get_comoving_radial_distance(self, z):
        r"""
        Returns the comoving radial distance to the given redshifts in Mpc.

        The redshifts must be a subset of those requested when
        :func:`~BoltzmannBase.must_provide` was called.
        """
        return self._get_z_dependent("comoving_radial_distance", z)

    def get_Pk_grid(self, var_pair=("delta_tot", "delta_tot"), nonlinear=True):
        """
        Get  matter power spectrum, e.g. suitable for splining.
        Returned arrays may be bigger or more densely sampled than requested, but will
        include required values. Neither k nor PK are in h^{-1} units.
        z and k are in ascending order.

        :param nonlinear: whether the linear or nonlinear spectrum
        :param var_pair: which power spectrum
        :return: k, z, PK, where k and z are arrays,
                 and PK[i,j] is the value at z[i], k[j]
        """
        try:
            return self._current_state[
                ("Pk_grid", bool(nonlinear)) + tuple(sorted(var_pair))]
        except KeyError:
            if ("Pk_grid", False) + tuple(sorted(var_pair)) in self._current_state:
                raise LoggedError(self.log,
                                  "Getting non-linear matter power but nonlinear "
                                  "not specified in requirements")
            raise LoggedError(self.log, "Matter power %s, %s not computed" % var_pair)

    def get_Pk_interpolator(self, var_pair=("delta_tot", "delta_tot"), nonlinear=True,
                            extrap_kmax=None):
        """
        Get P(z,k) bicubic interpolation object (:class:`PowerSpectrumInterpolator`).
        Neither k nor PK are in h^{-1} units.

        :param var_pair: variable pair for power spectrum
        :param nonlinear: non-linear spectrum (default True)
        :param extrap_kmax: use log linear extrapolation beyond max k computed up to
                            extrap_kmax
        :return: :class:`PowerSpectrumInterpolator` instance.
        """
        nonlinear = bool(nonlinear)
        key = ("Pk_interpolator", nonlinear, extrap_kmax) + tuple(sorted(var_pair))
        if key in self._current_state:
            return self._current_state[key]
        k, z, pk = self.get_Pk_grid(var_pair=var_pair, nonlinear=nonlinear)
        log_p = True
        sign = 1
        if np.any(pk < 0):
            if np.all(pk < 0):
                sign = -1
            else:
                log_p = False
        if log_p:
            pk = np.log(sign * pk)
        elif extrap_kmax > k[-1]:
            raise LoggedError(self.log,
                              'Cannot do log extrapolation with zero-crossing pk '
                              'for %s, %s' % var_pair)
        result = PowerSpectrumInterpolator(z, k, pk, logP=log_p, logsign=sign,
                                           extrap_kmax=extrap_kmax)
        self._current_state[key] = result
        return result

    def get_sigma_R(self, var_pair=("delta_tot", "delta_tot")):
        """
        Get sigma(R), the RMS power in an sphere of radius R
        Note R is in Mpc not h^{-1}Mpc units and z and R are returned in ascending order.

        You may get back more values than originally requested, but requested R and z
        should in the returned arrays.

        :param var_pair: which two fields to use for the RMS power
        :return: R, z, sigma_R, where R and z are arrays of computed values,
                 and sigma_R[i,j] is the value for z[i], R[j]
        """
        try:
            return self._current_state[("sigma_R",) + tuple(sorted(var_pair))]
        except KeyError:
            raise LoggedError(self.log, "sigmaR %s not computed" % var_pair)

    def get_source_Cl(self):
        r"""
        Returns a dict of power spectra of for the computed sources, with keys a tuple of
        sources ``([source1], [source2])``, and an additional key ``ell`` containing the
        multipoles.
        """

    def get_fsigma8(self, z):
        r"""
        Structure growth rate :math:`f\sigma_8`, as defined in eq. 33 of
        `Planck 2015 results. XIII. Cosmological parameters <https://arxiv.org/pdf/1502.01589.pdf>`_,
        at the given redshifts.

        The redshifts must be a subset of those requested when
        :func:`~BoltzmannBase.must_provide` was called.
        """
        pass

    def get_auto_covmat(self, params_info, likes_info):
        r"""
        Tries to get match to a database of existing covariance matrix files for the
        current model and data.

        ``params_info`` should contain preferably the slow parameters only.
        """
        from cobaya.cosmo_input import _get_best_covmat
        return _get_best_covmat(self.packages_path, params_info, likes_info)


class PowerSpectrumInterpolator(RectBivariateSpline):
    r"""
    2D spline interpolation object (scipy.interpolate.RectBivariateSpline)
    to evaluate matter power spectrum as function of z and k.

    *This class is adapted from CAMB's own P(k) interpolator, by Antony Lewis;
    it's mostly interface-compatible with the original.*

    :param z: values of z for which the power spectrum was evaluated.
    :param k: values of k for which the power spectrum was evaluated.
    :param P_or_logP: Values of the power spectrum (or log-values, if logP=True).
    :param logP: if True (default: False), log of power spectrum are given and used
        for the underlying interpolator.
    :param logsign: if logP is True, P_or_logP is log(logsign*Pk)
    :param extrap_kmax: if set, use power law extrapolation beyond kmax up to
        extrap_kmax; useful for tails of integrals.
    """

    def __init__(self, z, k, P_or_logP, extrap_kmax=None, logP=False, logsign=1):
        self.islog = logP
        #  Check order
        z, k = (np.atleast_1d(x) for x in [z, k])
        if len(z) < 4:
            raise ValueError('Require at least four redshifts for Pk interpolation.'
                             'Consider using Pk_grid if you just need a a small number'
                             'of specific redshifts (doing 1D splines in k yourself).')
        i_z = np.argsort(z)
        i_k = np.argsort(k)
        self.logsign = logsign
        self.z, self.k, P_or_logP = z[i_z], k[i_k], P_or_logP[i_z, :][:, i_k]
        self.zmin, self.zmax = self.z[0], self.z[-1]
        self.kmin, self.kmax = self.k[0], self.k[-1]
        logk = np.log(self.k)
        # Continue until extrap_kmax using a (log,log)-linear extrapolation
        if extrap_kmax and extrap_kmax > self.kmax:
            if not logP:
                raise ValueError('extrap_kmax must use logP')
            logk = np.hstack(
                [logk, np.log(self.kmax) * 0.1 + np.log(extrap_kmax) * 0.9,
                 np.log(extrap_kmax)])
            logPnew = np.empty((P_or_logP.shape[0], P_or_logP.shape[1] + 2))
            logPnew[:, :-2] = P_or_logP
            diff = (logPnew[:, -3] - logPnew[:, -4]) / (logk[-3] - logk[-4])
            delta = diff * (logk[-1] - logk[-3])
            logPnew[:, -1] = logPnew[:, -3] + delta
            logPnew[:, -2] = logPnew[:, -3] + delta * 0.9
            self.kmax = extrap_kmax  # Added for consistency with CAMB

            P_or_logP = logPnew

        super().__init__(self.z, logk, P_or_logP)

    def P(self, z, k, grid=None):
        """
        Get the power spectrum at (z,k).
        """
        if grid is None:
            grid = not np.isscalar(z) and not np.isscalar(k)
        if self.islog:
            return self.logsign * np.exp(self(z, np.log(k), grid=grid))
        else:
            return self(z, np.log(k), grid=grid)

    def logP(self, z, k, grid=None):
        """
        Get the log power spectrum at (z,k). (or minus log power spectrum if
        islog and logsign=-1)
        """
        if grid is None:
            grid = not np.isscalar(z) and not np.isscalar(k)
        if self.islog:
            return self(z, np.log(k), grid=grid)
        else:
            return np.log(self(z, np.log(k), grid=grid))



"""
.. module:: parameterization

:Synopsis: Class managing the possibly different parameterizations
           used by sampler and likelihoods
:Author: Jesus Torrado

"""
# Global
import numpy as np
from numbers import Number
from itertools import chain
from copy import deepcopy

# Local
from cobaya.conventions import partag
from cobaya.tools import get_external_function, ensure_nolatex, is_valid_variable_name
from cobaya.tools import getfullargspec, deepcopy_where_possible
from cobaya.log import LoggedError, HasLogger


def is_fixed_param(info_param):
    """
    Returns True if the parameter has been fixed to a value or through a function.
    """
    return expand_info_param(info_param).get(partag.value, None) is not None


def is_sampled_param(info_param):
    """
    Returns True if the parameter has a prior.
    """
    return partag.prior in expand_info_param(info_param)


def is_derived_param(info_param):
    """
    Returns True if the parameter is saved as a derived one.
    """
    return expand_info_param(info_param).get(partag.derived, False)


def expand_info_param(info_param, default_derived=True):
    """
    Expands the info of a parameter, from the user friendly, shorter format
    to a more unambiguous one.
    """
    info_param = deepcopy_where_possible(info_param)
    if not isinstance(info_param, dict):
        if info_param is None:
            info_param = {}
        else:
            info_param = {partag.value: info_param}
    if all(f not in info_param for f in [partag.prior, partag.value, partag.derived]):
        info_param[partag.derived] = default_derived
    # Dynamical input parameters: save as derived by default
    value = info_param.get(partag.value, None)
    if isinstance(value, str) or callable(value):
        info_param[partag.derived] = info_param.get(partag.derived, True)
    return info_param


def reduce_info_param(info_param):
    """
    Compresses the info of a parameter, suppressing default values.
    This is the opposite of :func:`~input.expand_info_param`.
    """
    info_param = deepcopy_where_possible(info_param)
    if not isinstance(info_param, dict):
        return
    # All parameters without a prior are derived parameters unless otherwise specified
    if info_param.get(partag.derived) is True:
        info_param.pop(partag.derived)
    # Fixed parameters with single "value" key
    if list(info_param) == [partag.value]:
        return info_param[partag.value]
    return info_param


class Parameterization(HasLogger):
    """
    Class managing parameterization.
    Translates parameter between sampler+prior and likelihood
    """

    def __init__(self, info_params, allow_renames=True, ignore_unused_sampled=False):
        self.set_logger(lowercase=True)
        self.allow_renames = allow_renames
        # First, we load the parameters,
        # not caring about whether they are understood by any likelihood.
        # `input` contains the parameters (expected to be) understood by the likelihood,
        #   with its fixed value, its fixing function, or None if their value is given
        #   directly by the sampler.
        self._infos = {}
        self._input = {}
        self._input_funcs = {}
        self._input_args = {}
        self._output = {}
        self._constant = {}
        self._sampled = {}
        self._sampled_renames = {}
        self._derived = {}
        self._derived_funcs = {}
        self._derived_args = {}
        # Notice here that expand_info_param *always* adds a partag.derived:True tag
        # to infos without _prior or partag.value, and a partag.value field
        # to fixed params
        for p, info in info_params.items():
            self._infos[p] = deepcopy_where_possible(info)
            if is_fixed_param(info):
                if isinstance(info[partag.value], Number):
                    self._constant[p] = info[partag.value]
                    if not info.get(partag.drop, False):
                        self._input[p] = self._constant[p]
                else:
                    self._input[p] = None
                    self._input_funcs[p] = get_external_function(info[partag.value])
                    self._input_args[p] = getfullargspec(self._input_funcs[p]).args
            if is_sampled_param(info):
                self._sampled[p] = None
                if not info.get(partag.drop, False):
                    self._input[p] = None
                self._sampled_renames[p] = (
                    (lambda x: [x] if isinstance(x, str) else x)
                    (info.get(partag.renames, [])))
            if is_derived_param(info):
                self._derived[p] = deepcopy_where_possible(info)
                # Dynamical parameters whose value we want to save
                if info[partag.derived] is True and is_fixed_param(info):
                    info[partag.derived] = "lambda %s: %s" % (p, p)
                if info[partag.derived] is True:
                    self._output[p] = None
                else:
                    self._derived_funcs[p] = get_external_function(info[partag.derived])
                    self._derived_args[p] = getfullargspec(self._derived_funcs[p]).args
        # Check that the sampled and derived params are all valid python variable names
        for p in chain(self._sampled, self._derived):
            if not is_valid_variable_name(p):
                is_in = p in self._sampled
                eg_in = "  p_prime:\n    prior: ...\n  %s: 'lambda p_prime: p_prime'\n" % p
                eg_out = "  p_prime: 'lambda %s: %s'\n" % (p, p)
                raise LoggedError(
                    self.log, "Parameter name '%s' is not a valid Python variable name "
                              "(it needs to start with a letter or '_').\n"
                              "If this is an %s parameter of a likelihood or theory, "
                              "whose name you cannot change,%s define an associated "
                              "%s one with a valid name 'p_prime' as: \n\n%s",
                    p, "input" if is_in else "output",
                    "" if is_in else " remove it and",
                    "sampled" if is_in else "derived",
                    eg_in if is_in else eg_out)
        # Assume that the *un*known function arguments are likelihood/theory
        # output parameters
        for arg in (set(chain(*self._input_args.values()))
                            .union(chain(*self._derived_args.values()))
                    - set(self._constant) - set(self._input)
                    - set(self._sampled) - set(self._derived)):
            self._output[arg] = None

        # Useful sets: directly-sampled input parameters and directly "output-ed" derived
        self._directly_sampled = [p for p in self._input if p in self._sampled]
        self._directly_output = [p for p in self._derived if p in self._output]
        # Useful mapping: input params that vary if each sample is varied
        self._sampled_input_dependence = {s: [i for i in self._input
                                              if s in self._input_args.get(i, {})]
                                          for s in self._sampled}
        # From here on, some error control.
        dropped_but_never_used = (
            set(p for p, v in self._sampled_input_dependence.items() if not v)
                .difference(set(self._directly_sampled)))
        if dropped_but_never_used and not ignore_unused_sampled:
            raise LoggedError(
                self.log,
                "Parameters %r are sampled but not passed to a likelihood or theory "
                "code, and never used as arguments for any parameter functions. "
                "Check that you are not using the '%s' tag unintentionally.",
                list(dropped_but_never_used), partag.drop)
        # input params depend on input and sampled only, never on output/derived
        all_input_arguments = set(chain(*self._input_args.values()))
        bad_input_dependencies = all_input_arguments.difference(
            set(self.input_params()).union(set(self.sampled_params())).union(
                set(self.constant_params())))
        if bad_input_dependencies:
            raise LoggedError(
                self.log,
                "Input parameters defined as functions can only depend on other "
                "input parameters that are not defined as functions. "
                "In particular, an input parameter cannot depend on %r."
                "Use an explicit Theory calculator for more complex dependencies.",
                list(bad_input_dependencies))
        self._wrapped_input_funcs, self._wrapped_derived_funcs = \
            self._get_wrapped_functions_evaluation_order()

    def input_params(self):
        return self._input.copy()

    def output_params(self):
        return self._output.copy()

    def constant_params(self):
        return self._constant.copy()

    def sampled_params(self):
        return self._sampled.copy()

    def sampled_params_info(self):
        return {p: deepcopy_where_possible(info) for p, info
                in self._infos.items() if p in self._sampled}

    def sampled_params_renames(self):
        return deepcopy(self._sampled_renames)

    def derived_params(self):
        return self._derived.copy()

    def sampled_input_dependence(self):
        return deepcopy(self._sampled_input_dependence)

    def to_input(self, sampled_params_values, copied=True):
        # Store sampled params, so that derived can depend on them
        if not isinstance(sampled_params_values, dict):
            sampled_params_values = dict(zip(self._sampled, sampled_params_values))
        else:
            sampled_params_values = sampled_params_values.copy()

        self._sampled = sampled_params_values
        # Fill first directly sampled input parameters
        for p in self._directly_sampled:
            self._input[p] = sampled_params_values[p]
        if self._wrapped_input_funcs:
            # Then evaluate the functions
            for p, (func, args, to_set) in self._wrapped_input_funcs.items():
                for arg in to_set:
                    args[arg] = self._input.get(arg,
                                                sampled_params_values.get(arg, None))
                self._input[p] = self._call_param_func(p, func, args)
        return self.input_params() if copied else self._input

    def to_derived(self, output_params_values):
        if not isinstance(output_params_values, dict):
            output_params_values = dict(zip(self._output, output_params_values))
        # Fill first derived parameters which are direct output parameters
        for p in self._directly_output:
            self._derived[p] = output_params_values[p]
        # Then evaluate the functions
        if self._wrapped_derived_funcs:
            # Then evaluate the functions
            for p, (func, args, to_set) in self._wrapped_derived_funcs.items():
                for arg in to_set:
                    val = self._input.get(arg)
                    if val is None:
                        val = output_params_values.get(arg)
                        if val is None:
                            val = self._derived.get(arg)
                            if val is None:
                                val = self._sampled.get(arg)
                    args[arg] = val
                self._derived[p] = self._call_param_func(p, func, args)

        return list(self._derived.values())

    def check_sampled(self, **sampled_params):
        """
        Check that the input dictionary contains all the sampled parameters,
        and just them. Is aware of known renamings.

        Returns dict of parameters (model's naming) and their values.
        """
        sampled_output = {}
        sampled_input = sampled_params.copy()
        for p, renames in self._sampled_renames.items():
            for pprime in sampled_input:
                if pprime == p or (pprime in renames if self.allow_renames else False):
                    sampled_output[p] = sampled_input.pop(pprime)
                    break
        if len(sampled_output) < len(self._sampled):
            not_found = set(self.sampled_params()).difference(set(sampled_output))
            if self.allow_renames:
                msg = ("The following expected sampled parameters " +
                       ("(or their aliases) " if self.allow_renames else "") +
                       "where not found : %r",
                       ({p: self._sampled_renames[p] for p in not_found}
                        if self.allow_renames else not_found))
            else:
                msg = ("The following expected sampled parameters "
                       "where not found : %r",
                       {p: self._sampled_renames[p] for p in not_found})
            raise LoggedError(self.log, *msg)
        # Ignore fixed input parameters if they have the correct value
        to_pop = []
        for p, value in sampled_input.items():
            known_value = self.constant_params().get(p, None)
            if known_value is None:
                raise LoggedError(self.log, "Unknown parameter %r.", p)
            elif np.allclose(value, known_value):
                to_pop.append(p)
                self.log.debug("Fixed parameter %r ignored.", p)
            else:
                raise LoggedError(
                    self.log, "Cannot change value of constant parameter: "
                              "%s = %g (new) vs %g (old).", p, value, known_value)
        for p in to_pop:
            sampled_input.pop(p)
        if sampled_input:
            not_used = set(sampled_input)
            duplicated = not_used.intersection(set(
                chain(
                    *[list(chain(*[[k], v])) for k, v in self._sampled_renames.items()])))
            not_used = not_used.difference(duplicated)
            derived = not_used.intersection(set(self.derived_params()))
            input_ = not_used.intersection(set(self.input_params()))
            unknown = not_used.difference(derived).difference(input_)
            msg = ("Incorrect parameters! " +
                   ("\n   Duplicated entries (using their aliases): %r" % list(duplicated)
                    if duplicated else "") +
                   ("\n   Not known: %r" % list(unknown) if unknown else "") +
                   ("\n   Cannot be fixed: %r " % list(input_) +
                    "--> instead, fix sampled parameters that depend on them!"
                    if input_ else "") +
                   ("\n   Cannot be fixed because are derived parameters: %r " % list(
                       derived) if derived else ""))
            for line in msg.split("\n"):
                self.log.error(line)
            raise LoggedError
        return sampled_output

    def labels(self):
        """
        Returns a dictionary of LaTeX labels of the sampled and derived parameters.

        Uses the parameter name if no label has been given.
        """

        def get_label(p, info):
            return ensure_nolatex(getattr(info, "get", lambda x, y: y)
                                  (partag.latex, p.replace("_", r"\ ")))

        return {p: get_label(p, info) for p, info in self._infos.items()}

    def _call_param_func(self, p, func, kwargs):
        try:
            return func(**kwargs)
        except NameError as exception:
            unknown = str(exception).split("'")[1]
            raise LoggedError(
                self.log, "Unknown variable '%s' was referenced in the definition of "
                          "the parameter '%s', with arguments %r.", unknown, p,
                list(kwargs))
        except:
            self.log.error("Function for parameter '%s' failed at evaluation "
                           "and threw the following exception:", p)
            raise

    def _get_wrapped_functions_evaluation_order(self):
        # get evaluation order for input and derived parameter function
        # and pre-prepare argument dicts

        wrapped_funcs = ({}, {})
        known = set(self._constant).union(self._sampled)

        for derived, wrapped_func in zip((False, True), wrapped_funcs):
            if derived:
                inputs = self._derived_funcs.copy()
                input_args = self._derived_args
                known.update(self._output)
                output = self._derived
            else:
                inputs = self._input_funcs.copy()
                input_args = self._input_args
                output = self._input

            while inputs:
                for p, func in inputs.items():
                    args = input_args[p]
                    if set(args).difference(known):
                        continue
                    known.add(p)

                    if not set(args).difference(self._constant):
                        # all inputs are constant, so output is constant and precomputed
                        self._constant[p] = \
                            self._call_param_func(p, func,
                                                  {arg: self._constant[arg] for arg in
                                                   args})
                        output[p] = self._constant[p]
                    else:
                        # Store function, argument dict with constants pre-filled,
                        # and unset args as tuple
                        wrapped_func[p] = \
                            (func, {arg: self._constant.get(arg) for arg in args},
                             [arg for arg in args if arg not in self._constant])

                    del inputs[p]
                    break
                else:
                    raise LoggedError(
                        self.log, "Could not resolve arguments for parameters %s. "
                                  "Maybe there is a circular dependency between derived "
                                  "parameters?", list(inputs))
        return wrapped_funcs

    # Python magic for the "with" statement
    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        return



r"""
.. module:: _H0_prototype

:Synopsis: Prototype class for local Hubble parameter measurements
:Author: Jesus Torrado

This is a simple gaussian likelihood for the latest local :math:`H_0` measurements
using a combination of different data.

It defines the following likelihoods

- ``H0.riess2018a``: A legacy local measurement of :math:`H_0`, used in the analysis of
  Planck data for the Planck 2018 release.
  `(arXiv:1801.01120) <https://arxiv.org/abs/1801.01120>`_
- ``H0.riess2018b``: Updated local measurement of :math:`H_0`.
  `(arXiv:1804.10655) <https://arxiv.org/abs/1804.10655>`_
- ``H0.riess201903``:  Riess et al. 2019 constraint
  `(arXiv:1903.07603) <https://arxiv.org/abs/1903.07603>`_


Using a different measurement
-----------------------------

If you would like to use different values for the :math:`H_0` constraint, as a mean and a
standard deviation, simply add the following likelihood, substituting ``mu_H0`` and
``sigma_H0`` for the respective values:

.. literalinclude:: ./src_examples/H0/custom_likelihood.yaml
   :language: yaml

"""

# Global
from scipy.stats import norm

# Local
from cobaya.likelihood import Likelihood


class _H0_prototype(Likelihood):
    # Data type for aggregated chi2 (case sensitive)
    type = "H0"

    # variables from yaml
    H0_mean: float
    H0_std: float

    def initialize(self):
        self.norm = norm(loc=self.H0_mean, scale=self.H0_std)

    def get_requirements(self):
        return {'H0': None}

    def logp(self, **params_values):
        H0_theory = self.provider.get_param("H0")
        return self.norm.logpdf(H0_theory)



"""
.. module:: conventions

:Synopsis: Some physical constants and naming conventions
           to make the life of the maintainer easier.
:Author: Jesus Torrado

"""
from collections import namedtuple
from types import MappingProxyType

# Package name (for importlib)
# (apparently __package__ is only defined if you import something locally.
_cobaya_package = __name__.rpartition('.')[0]

# an immutable empty dict (e.g. for argument defaults)
empty_dict = MappingProxyType({})

# Names for block and fields in the input
_prior = "prior"
_post = "post"
_post_add = "add"
_post_remove = "remove"
_post_suffix = "suffix"
_params = "params"
_auto_params = "auto_params"
_input_params = "input_params"
_output_params = "output_params"
_input_params_prefix = "input_params_prefix"
_output_params_prefix = "output_params_prefix"
_debug = "debug"
_debug_default = False
_debug_file = "debug_file"
_output_prefix = "output"
_packages_path = "packages_path"
_external = "external"
_provides = "provides"
_requires = "requires"
_resume = "resume"
_resume_default = False
_timing = "timing"
_force = "force"
_test_run = "test"
_component_path = "python_path"
_component_class_name = "class_name"
_aliases = "aliases"
_version = "version"

ParTags = namedtuple('ParTags', ("prior", "ref", "proposal", "value", "dist", "drop",
                                 "derived", "latex", "renames"))
partag = ParTags(*ParTags._fields)

ComponentKinds = namedtuple('ComponentKinds', ("sampler", "theory", "likelihood"))
kinds = ComponentKinds(*ComponentKinds._fields)

reserved_attributes = {_input_params, _output_params, "install_options"}

# Conventional order for yaml dumping (purely cosmetic)
_dump_sort_cosmetic = [
    kinds.theory, kinds.likelihood, _prior, _params, kinds.sampler, "post"]

# Separator for
# fields in parameter names and files
# Its manual inclusion in a string anywhere else (e.g. a parameter name) should be avoided
_separator = "__"
_separator_files = "."

# Names for the samples' fields internally and in the output
_weight = "weight"  # sample weight
_minuslogpost = "minuslogpost"  # log-posterior, or in general the total log-probability
_minuslogprior = "minuslogprior"  # log-prior
_chi2 = "chi2"  # chi^2 = -2 * loglik
_get_chi2_name = lambda p: _chi2 + _separator + str(p)
_undo_chi2_name = lambda p: p[len(_chi2 + _separator):]
_get_chi2_label = lambda p: r"\chi^2_\mathrm{" + str(p).replace("_", r"\ ") + r"}"
_prior_1d_name = "0"

# Output files
_input_suffix = "input"
_updated_suffix = "updated"
_yaml_extensions = ".yaml", ".yml"
_checkpoint_extension = ".checkpoint"
_progress_extension = ".progress"
_covmat_extension = ".covmat"
_evidence_extension = ".logZ"

# Installation and container definitions
_packages_path_arg = _packages_path
_packages_path_arg_posix = _packages_path_arg.replace("_", "-")
_packages_path_env = "COBAYA_PACKAGES_PATH"
_packages_path_config_file = "config.yaml"
_packages_path_containers = "/cobaya_packages"
_code = "code"
_data = "data"
_install_skip_env = "COBAYA_INSTALL_SKIP"
_test_skip_env = "COBAYA_TEST_SKIP"
_covmats_file = "covmats_database.pkl"
_products_path = "/products"
_requirements_file = "requirements.yaml"
_help_file = "readme.md"

# Internal package structure
subfolders = {kinds.likelihood: "likelihoods",
              kinds.sampler: "samplers",
              kinds.theory: "theories"}

# Approximate overhead of cobaya per posterior evaluation. Useful for blocking speeds
_overhead_time = 0.0003

# Line width for console printing
_line_width = 120

# Physical constants
# ------------------
# Light speed
_c_km_s = 299792.458  # speed of light
_h_J_s = 6.626070040e-34  # Planck's constant
_kB_J_K = 1.38064852e-23  # Boltzmann constant



import sys
import platform

if sys.version_info[0] == 2:
    print('Cobaya no longer supports Python 2, please upgrade to Python 3')
    sys.exit(1)

if sys.version_info < (3, 7):
    if sys.version_info < (3, 6):
        print('Cobaya requires Python 3.6+, please upgrade.')
        sys.exit(1)

    if platform.python_implementation() != 'CPython':
        raise ValueError('Cobaya only supports CPython implementations on Python 3.6')

__author__ = "Jesus Torrado and Antony Lewis"
__version__ = "3.0"
__obsolete__ = False
__year__ = "2020"
__url__ = "https://cobaya.readthedocs.io"



"""
.. module:: theories.camb

:Synopsis: Managing the CAMB cosmological code
:Author: Jesus Torrado and Antony Lewis

.. |br| raw:: html

   <br />

This module imports and manages the CAMB cosmological code.
It requires CAMB 1.1.2 or higher.

.. note::

   **If you use this cosmological code, please cite it as:**
   |br|
   A. Lewis, A. Challinor, A. Lasenby,
   *Efficient computation of CMB anisotropies in closed FRW*
   (`arXiv:astro-ph/9911177 <https://arxiv.org/abs/astro-ph/9911177>`_)
   |br|
   C. Howlett, A. Lewis, A. Hall, A. Challinor,
   *CMB power spectrum parameter degeneracies in the era of precision cosmology*
   (`arXiv:1201.3654 <https://arxiv.org/abs/1201.3654>`_)


Usage
-----

If you are using a likelihood that requires some observable from CAMB, simply add CAMB
to the theory block.

You can specify any parameter that CAMB understands in the ``params`` block:

.. code-block:: yaml

   theory:
     camb:
       extra_args:
         [any param that CAMB understands, for FIXED and PRECISION]

   params:
       [any param that CAMB understands, fixed, sampled or derived]


If you want to use your own version of CAMB, you need to specify its location with a
``path`` option inside the ``camb`` block. If you do not specify a ``path``,
CAMB will be loaded from the automatic-install ``packages_path`` folder, if specified, or
otherwise imported as a globally-installed Python package. Cobaya will print at
initialisation where it is getting CAMB from.

.. _camb_modify:

Modifying CAMB
^^^^^^^^^^^^^^

If you modify CAMB and add new variables, make sure that the variables you create are
exposed in the Python interface (`instructions here
<https://camb.readthedocs.io/en/latest/model.html#camb.model.CAMBparams>`__).
If you follow those instructions you do not need to make any additional modification in
Cobaya.

You can use the :doc:`model wrapper <cosmo_model>` to test your modification by
evaluating observables or getting derived quantities at known points in the parameter
space (set ``debug: True`` to get more detailed information of what exactly is passed to
CAMB).

In your CAMB modification, remember that you can raise a ``CAMBParamRangeError`` or a
``CAMBError`` whenever the computation of any observable would fail, but you do not
expect that observable to be compatible with the data (e.g. at the fringes of the
parameter space). Whenever such an error is raised during sampling, the likelihood is
assumed to be zero, and the run is not interrupted.


Installation
------------

Pre-requisites
^^^^^^^^^^^^^^

**cobaya** calls CAMB using its Python interface, which requires that you compile CAMB
using intel's ifort compiler or the GNU gfortran compiler version 6.4 or later.
To check if you have the latter, type ``gfortran --version`` in the shell,
and the first line should look like

.. code::

   GNU Fortran ([your OS version]) [gfortran version] [release date]

Check that ``[gfortran's version]`` is at least 6.4. If you get an error instead, you need
to install gfortran (contact your local IT service).

CAMB comes with binaries pre-built for Windows, so if you don't need to modify the CAMB
source code, no Fortran compiler is needed.

If you are using Anaconda you can also install a pre-compiled CAMB package from conda
forge using

.. code::

  conda install -c conda-forge camb

Automatic installation
^^^^^^^^^^^^^^^^^^^^^^

If you do not plan to modify CAMB, the easiest way to install it is using the
:doc:`automatic installation script <installation_cosmo>`. Just make sure that
``theory: camb:`` appears in one of the files passed as arguments to the installation
script.


Manual installation (or using your own version)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

If you are planning to modify CAMB or use an already modified version,
you should not use the automatic installation script. Use the installation method that
best adapts to your needs:

* [**Recommended for staying up-to-date**]
  To install CAMB locally and keep it up-to-date, clone the
  `CAMB repository in Github <https://github.com/cmbant/CAMB>`_
  in some folder of your choice, say ``/path/to/theories/CAMB``:

  .. code:: bash

      $ cd /path/to/theories
      $ git clone --recursive https://github.com/cmbant/CAMB.git
      $ cd CAMB
      $ python setup.py build

  To update to the last changes in CAMB (master), run ``git pull`` from CAMB's folder and
  re-build using the last command. If you do not want to use multiple versions of CAMB,
  you can also make your local installation available to python generally by installing
  it using

.. code:: bash

     $ python -m pip install -e /path/to/CAMB

* [**Recommended for modifying CAMB**]
  First, `fork the CAMB repository in Github <https://github.com/cmbant/CAMB>`_
  (follow `these instructions <https://help.github.com/articles/fork-a-repo/>`_) and then
  follow the same steps as above, substituting the second one with:

  .. code:: bash

      $ git clone --recursive https://[YourGithubUser]@github.com/[YourGithubUser]/CAMB.git

* To use your own version, assuming it's placed under ``/path/to/theories/CAMB``,
  just make sure it is compiled (and that the version on top of which you based your
  modifications is old enough to have the Python interface implemented.

In the cases above, you **must** specify the path to your CAMB installation in
the input block for CAMB (otherwise a system-wide CAMB may be used instead):

.. code:: yaml

   theory:
     camb:
       path: /path/to/theories/CAMB

.. note::

   In any of these methods, if you intent to switch between different versions or
   modifications of CAMB you should not  install CAMB as python package using
   ``python setup.py install --user``, as the official instructions suggest.
"""

# Global
import sys
import os
import logging
from copy import deepcopy
import numpy as np
import numbers
import ctypes
from typing import NamedTuple, Any

# Local
from cobaya.theories._cosmo import BoltzmannBase
from cobaya.log import LoggedError
from cobaya.install import download_github_release, check_gcc_version
from cobaya.tools import getfullargspec, get_class_methods, get_properties
from cobaya.tools import load_module, VersionCheckError, str_to_list
from cobaya.theory import HelperTheory
from cobaya.conventions import _requires


# Result collector
class Collector(NamedTuple):
    method: callable
    args: list = []
    kwargs: dict = {}


class CAMBOutputs(NamedTuple):
    camb_params: Any
    results: Any
    derived: dict


class camb(BoltzmannBase):
    # Name of the Class repo/folder and version to download
    _camb_repo_name = "cmbant/CAMB"
    _camb_repo_version = os.environ.get("CAMB_REPO_VERSION", "master")
    _camb_min_gcc_version = "6.4"
    _min_camb_version = '1.1.2'

    external_primordial_pk: bool

    def initialize(self):
        """Importing CAMB from the correct path, if given."""
        if not self.path and self.packages_path:
            self.path = self.get_path(self.packages_path)
        camb_path = None
        if self.path and not os.path.exists(self.path):
            # Fail if this was a directly specified path,
            # or ignore and try to global-import if it came from a packages_path
            if self.packages_path:
                self.log.info("*local* CAMB not found at " + self.path)
                self.log.info("Importing *global* CAMB.")
            else:
                raise LoggedError(self.log, "*local* CAMB not found at " + self.path)
        elif self.path:
            self.log.info("Importing *local* CAMB from " + self.path)
            if not os.path.exists(self.path):
                raise LoggedError(
                    self.log, "The given folder does not exist: '%s'", self.path)
            camb_path = self.path
            if not os.path.exists(os.path.join(self.path, "setup.py")):
                raise LoggedError(
                    self.log,
                    "Either CAMB is not in the given folder, '%s', or you are using a "
                    "very old version without the Python interface.", self.path)
        else:
            self.log.info("Importing *global* CAMB.")
        try:
            self.camb = load_module("camb", path=camb_path,
                                    min_version=self._min_camb_version)
        except ImportError:
            raise LoggedError(
                self.log, "Couldn't find the CAMB python interface.\n"
                          "Make sure that you have compiled it, and that you either\n"
                          " (a) specify a path (you didn't) or\n"
                          " (b) install the Python interface globally with\n"
                          "     'python -m pip install -e /path/to/camb [--user]'")
        except VersionCheckError as e:
            raise LoggedError(self.log, str(e))
        super().initialize()
        self.extra_attrs = {"Want_CMB": False, "Want_cl_2D_array": False,
                            'WantCls': False}
        # Derived parameters that may not have been requested, but will be necessary later
        self.derived_extra = []
        # Some default settings
        self.needs_perts = False
        self.limber = False
        self.non_linear_sources = False
        self.non_linear_pk = False
        self._base_params = None
        self._needs_lensing_cross = False
        self._sigmaR_z_indices = {}

        if self.external_primordial_pk:
            self.extra_args['initial_power_model'] \
                = self.camb.initialpower.SplinedInitialPower
            self.initial_power_args, self.power_params = {}, []
        else:
            power_spectrum = self.camb.CAMBparams.make_class_named(
                self.extra_args.get('initial_power_model',
                                    self.camb.initialpower.InitialPowerLaw),
                self.camb.initialpower.InitialPower)
            self.initial_power_args, self.power_params = \
                self._extract_params(power_spectrum.set_params)

        nonlin = self.camb.CAMBparams.make_class_named(
            self.extra_args.get('non_linear_model',
                                self.camb.nonlinear.Halofit),
            self.camb.nonlinear.NonLinearModel)

        self.nonlin_args, self.nonlin_params = self._extract_params(nonlin.set_params)

        self.requires = str_to_list(getattr(self, _requires, []))
        self._transfer_requires = [p for p in self.requires if
                                   p not in self.get_can_support_params()]
        self.requires = [p for p in self.requires if p not in self._transfer_requires]

    def _extract_params(self, set_func):
        args = {}
        params = []
        pars = getfullargspec(set_func)
        for arg, v in zip(pars.args[1:], pars.defaults[1:]):
            if arg in self.extra_args:
                args[arg] = self.extra_args.pop(arg)
            elif isinstance(v, numbers.Number) or v is None:
                params.append(arg)
        return args, params

    def initialize_with_params(self):
        # must set WantTensors manually if using external_primordial_pk
        if not self.external_primordial_pk \
                and set(self.input_params).intersection({'r', 'At'}):
            self.extra_attrs["WantTensors"] = True

    def get_can_support_params(self):
        return self.power_params + self.nonlin_params

    def get_allow_agnostic(self):
        return False

    def must_provide(self, **requirements):
        # Computed quantities required by the likelihoods
        # Note that redshifts below are treated differently for background quantities,
        #   where no additional transfer computation is needed (e.g. H(z)),
        #   and matter-power-related quantities, that require additional computation
        #   and need the redshifts to be passed at CAMBParams instantiation.
        #   Also, we always make sure that those redshifts are sorted in descending order,
        #   since all CAMB related functions return quantities in that implicit order
        # The following super call makes sure that the requirements are properly
        # accumulated, i.e. taking the max of precision requests, etc.
        super().must_provide(**requirements)
        CAMBdata = self.camb.CAMBdata

        for k, v in self._must_provide.items():
            # Products and other computations
            if k == "Cl":
                self.extra_args["lmax"] = max(
                    max(v.values()), self.extra_args.get("lmax", 0))
                cls = [a.lower() for a in v]
                needs_lensing = set(cls).intersection({"pp", "pt", "pe", "tp", "ep"})
                self.collectors[k] = Collector(
                    method=CAMBdata.get_cmb_power_spectra,
                    kwargs={
                        "spectra": list(set(
                            (self.collectors[k].kwargs.get("spectra", [])
                             if k in self.collectors else []) +
                            ["total"] + (["lens_potential"] if needs_lensing else []))),
                        "raw_cl": False})
                self.needs_perts = True
                self.extra_attrs["Want_CMB"] = True
                self.extra_attrs["WantCls"] = True
                if "pp" in cls and self.extra_args.get(
                        "lens_potential_accuracy") is None:
                    self.extra_args["lens_potential_accuracy"] = 1
                self.non_linear_sources = self.extra_args.get("lens_potential_accuracy",
                                                              1) >= 1
                if set(cls).intersection({"pt", "pe", "tp", "ep"}):
                    self._needs_lensing_cross = True
                if 'TCMB' not in self.derived_extra:
                    self.derived_extra += ['TCMB']
            elif k == "Hubble":
                self.collectors[k] = Collector(
                    method=CAMBdata.h_of_z,
                    kwargs={"z": self._combine_z(k, v)})
            elif k in ("angular_diameter_distance", "comoving_radial_distance"):
                self.collectors[k] = Collector(
                    method=getattr(CAMBdata, k),
                    kwargs={"z": self._combine_z(k, v)})
            elif k == "fsigma8":
                self.add_to_redshifts(v["z"])
                self.collectors[k] = Collector(
                    method=CAMBdata.get_fsigma8,
                    kwargs={})
                self.needs_perts = True
            elif isinstance(k, tuple) and k[0] == "sigma_R":
                kwargs = v.copy()
                self.extra_args["kmax"] = max(kwargs.pop("k_max"),
                                              self.extra_args.get("kmax", 0))
                redshifts = kwargs.pop("z")
                self.add_to_redshifts(redshifts)
                var_pair = k[1:]

                def get_sigmaR(results, **tmp):
                    _indices = self._sigmaR_z_indices.get(var_pair)
                    if not _indices:
                        z_indices = []
                        calc = np.array(results.Params.Transfer.PK_redshifts[
                                        :results.Params.Transfer.PK_num_redshifts])
                        for z in redshifts:
                            for i, zcalc in enumerate(calc):
                                if np.isclose(zcalc, z, rtol=1e-4):
                                    z_indices += [i]
                                    break
                            else:
                                raise LoggedError(self.log, "sigma_R redshift not found"
                                                            "in computed P_K array %s", z)
                        _indices = np.array(z_indices, dtype=np.int32)
                        self._sigmaR_z_indices[var_pair] = _indices
                    return results.get_sigmaR(hubble_units=False, return_R_z=True,
                                              z_indices=_indices, **tmp)

                kwargs.update(dict(zip(["var1", "var2"], var_pair)))
                self.collectors[k] = Collector(method=get_sigmaR, kwargs=kwargs)
                self.needs_perts = True
            elif isinstance(k, tuple) and k[0] == "Pk_grid":
                kwargs = v.copy()
                self.extra_args["kmax"] = max(kwargs.pop("k_max"),
                                              self.extra_args.get("kmax", 0))
                self.add_to_redshifts(kwargs.pop("z"))
                # need to ensure can't have conflicts between requests from
                # different likelihoods. Store results without Hubble units.
                if kwargs.get("hubble_units", False) or kwargs.get("k_hunit", False):
                    raise LoggedError(self.log, "hubble_units and k_hunit must be False"
                                                "for consistency")
                kwargs["hubble_units"] = False
                kwargs["k_hunit"] = False
                if kwargs["nonlinear"]:
                    self.non_linear_pk = True
                var_pair = k[2:]
                kwargs.update(dict(zip(["var1", "var2"], var_pair)))
                self.collectors[k] = Collector(
                    method=CAMBdata.get_linear_matter_power_spectrum,
                    kwargs=kwargs.copy())
                self.needs_perts = True
            elif k == "source_Cl":
                if not getattr(self, "sources", None):
                    self.sources = {}
                for source, window in v["sources"].items():
                    # If it was already there, BoltzmannBase.must_provide() has already
                    # checked that old info == new info
                    if source not in self.sources:
                        self.sources[source] = window
                self.limber = v.get("limber", True)
                self.non_linear_sources = self.non_linear_sources or \
                                          v.get("non_linear", False)
                if "lmax" in v:
                    self.extra_args["lmax"] = max(v["lmax"],
                                                  self.extra_args.get("lmax", 0))
                self.needs_perts = True
                self.collectors[k] = Collector(method=CAMBdata.get_source_cls_dict)
                self.extra_attrs["Want_cl_2D_array"] = True
                self.extra_attrs["WantCls"] = True
            elif k == 'CAMBdata':
                # Just get CAMB results object
                self.collectors[k] = None
            elif v is None:
                # General derived parameters
                k_translated = self.translate_param(k)
                if k_translated not in self.derived_extra:
                    self.derived_extra += [k_translated]
                if k == "sigma8":
                    self.extra_attrs["WantTransfer"] = True
                    self.needs_perts = True
            else:
                raise LoggedError(self.log, "This should not be happening. Contact the "
                                            "developers.")
        self.check_no_repeated_input_extra()

        # Computing non-linear corrections
        model = self.camb.model
        self.extra_attrs["NonLinear"] = {
            (True, True): model.NonLinear_both,
            (True, False): model.NonLinear_lens,
            (False, True): model.NonLinear_pk,
            (False, False): False}[(self.non_linear_sources, self.non_linear_pk)]
        # set-set base CAMB params if anything might have changed
        self._base_params = None

        must_provide = {'CAMB_transfers': {'non_linear': self.non_linear_sources,
                                           'needs_perts': self.needs_perts}}
        if self.external_primordial_pk and self.needs_perts:
            must_provide['primordial_scalar_pk'] = {'lmax': self.extra_args.get("lmax"),
                                                    'kmax': self.extra_args.get('kmax')}
            if self.extra_attrs.get('WantTensors'):
                must_provide['primordial_tensor_pk'] = {'lmax':
                    self.extra_attrs.get(
                        "max_l_tensor", self.extra_args.get("lmax"))}
        return must_provide

    def add_to_redshifts(self, z):
        self.extra_args["redshifts"] = np.sort(np.unique(np.concatenate(
            (np.atleast_1d(z), self.extra_args.get("redshifts", [])))))[::-1]

    def _combine_z(self, k, v):
        c = self.collectors.get(k, None)
        if c:
            return np.sort(
                np.unique(np.concatenate((c.kwargs['z'], np.atleast_1d(v['z'])))))
        else:
            return np.sort(np.atleast_1d(v['z']))

    def calculate(self, state, want_derived=True, **params_values_dict):
        try:
            params, results = self.provider.get_CAMB_transfers()
            if self.collectors:
                if self.external_primordial_pk and self.needs_perts:
                    primordial_pk = self.provider.get_primordial_scalar_pk()
                    if primordial_pk.get('log_regular', True):
                        results.Params.InitPower.set_scalar_log_regular(
                            primordial_pk['kmin'], primordial_pk['kmax'],
                            primordial_pk['Pk'])
                    else:
                        results.Params.InitPower.set_scalar_table(
                            primordial_pk['k'], primordial_pk['Pk']
                        )
                    results.Params.InitPower.effective_ns_for_nonlinear = \
                        primordial_pk.get('effective_ns_for_nonlinear', 0.97)
                    if self.extra_attrs.get("WantTensors"):
                        primordial_pk = self.provider.get_primordial_tensor_pk()
                        if primordial_pk.get('log_regular', True):
                            results.Params.InitPower.set_tensor_log_regular(
                                primordial_pk['kmin'], primordial_pk['kmax'],
                                primordial_pk['Pk'])
                        else:
                            results.Params.InitPower.set_tensor_table(
                                primordial_pk['k'], primordial_pk['Pk']
                            )
                else:
                    args = {self.translate_param(p): v for p, v in
                            params_values_dict.items() if p in self.power_params}
                    args.update(self.initial_power_args)
                    results.Params.InitPower.set_params(**args)
                if self.non_linear_sources or self.non_linear_pk:
                    args = {self.translate_param(p): v for p, v in
                            params_values_dict.items() if p in self.nonlin_params}
                    args.update(self.nonlin_args)
                    results.Params.NonLinearModel.set_params(**args)
                results.power_spectra_from_transfer()
            else:
                results = None
            for product, collector in self.collectors.items():
                if collector:
                    state[product] = \
                        collector.method(results, *collector.args, **collector.kwargs)
                else:
                    state[product] = results
        except self.camb.baseconfig.CAMBError as e:
            if self.stop_at_error:
                self.log.error(
                    "Computation error (see traceback below)! "
                    "Parameters sent to CAMB: %r and %r.\n"
                    "To ignore this kind of error, make 'stop_at_error: False'.",
                    dict(state["params"]), dict(self.extra_args))
                raise
            else:
                # Assumed to be a "parameter out of range" error.
                self.log.debug("Computation of cosmological products failed. "
                               "Assigning 0 likelihood and going on. "
                               "The output of the CAMB error was %s" % e)
                return False
            # Prepare derived parameters
        intermediates = CAMBOutputs(params, results,
                                    results.get_derived_params() if results else None)
        if want_derived:
            state["derived"] = self._get_derived_output(intermediates)
        # Prepare necessary extra derived parameters
        state["derived_extra"] = {
            p: self._get_derived(p, intermediates) for p in self.derived_extra}

    def _get_derived(self, p, intermediates):
        """
        General function to extract a single derived parameter.

        To get a parameter *from a likelihood* use `get_param` instead.
        """
        if intermediates.derived:
            derived = intermediates.derived.get(p, None)
            if derived is not None:
                return derived
        # Specific calls, if general ones fail:
        if p == "sigma8":
            return intermediates.results.get_sigma8()[-1]
        try:
            return getattr(intermediates.camb_params, p)
        except AttributeError:
            try:
                return getattr(intermediates.results, p)
            except AttributeError:
                return getattr(intermediates.camb_params, "get_" + p, lambda: None)()

    def _get_derived_output(self, intermediates):
        """
        Returns a dictionary of derived parameters with their values,
        using the *current* state (i.e. it should only be called from
        the ``compute`` method).

        To get a parameter *from a likelihood* use `get_param` instead.
        """
        derived = {}
        for p in self.output_params:
            derived[p] = self._get_derived(self.translate_param(p), intermediates)
            if derived[p] is None:
                raise LoggedError(self.log, "Derived param '%s' not implemented"
                                            " in the CAMB interface", p)
        return derived

    def get_Cl(self, ell_factor=False, units="muK2"):
        current_state = self._current_state
        # get C_l^XX from the cosmological code
        try:
            cl_camb = current_state["Cl"]["total"].copy()
        except:
            raise LoggedError(self.log, "No Cl's were computed. Are you sure that you "
                                        "have requested them?")

        temp = current_state['derived_extra']['TCMB']
        units_factors = {"1": 1,
                         "muK2": temp * 1.e6,
                         "K2": temp}
        try:
            units_factor = units_factors[units]
        except KeyError:
            raise LoggedError(self.log, "Units '%s' not recognized. Use one of %s.",
                              units, list(units_factors))

        ls = np.arange(cl_camb.shape[0], dtype=np.int64)
        if not ell_factor:
            # unit conversion and ell_factor. CAMB output is *with* the factors already
            ells_factor = ls[1:] * (ls[1:] + 1)
            cl_camb[1:, :] /= ells_factor[..., np.newaxis]
            cl_camb[1:, :] *= (2 * np.pi) * units_factor ** 2
        elif units_factor != 1:
            cl_camb *= units_factor ** 2

        mapping = {"tt": 0, "ee": 1, "bb": 2, "te": 3, "et": 3}
        cls = {"ell": ls}
        for sp, i in mapping.items():
            cls[sp] = cl_camb[:, i]

        cl_lens = current_state["Cl"].get("lens_potential")
        if cl_lens is not None:
            cls["pp"] = cl_lens[:, 0].copy()
            if not ell_factor:
                cls["pp"][1:] /= ells_factor ** 2 / (2 * np.pi)
            if self._needs_lensing_cross:
                for i, cross in enumerate(['pt', 'pe']):
                    cls[cross] = cl_lens[:, i + 1].copy() * units_factor
                    if not ell_factor:
                        cls[cross][1:] /= ells_factor ** (3. / 2) / (2 * np.pi)
                    cls[cross[::-1]] = cls[cross]
        return cls

    def _get_z_dependent(self, quantity, z):
        if quantity == "fsigma8":
            computed_redshifts = self.extra_args["redshifts"]
            i_kwarg_z = np.concatenate(
                [np.where(computed_redshifts == zi)[0] for zi in np.atleast_1d(z)])
        else:
            computed_redshifts = self.collectors[quantity].kwargs["z"]
            i_kwarg_z = np.searchsorted(computed_redshifts, np.atleast_1d(z))
        return np.array(self._current_state[quantity], copy=True)[i_kwarg_z]

    def get_fsigma8(self, z):
        return self._get_z_dependent("fsigma8", z)

    def get_source_Cl(self):
        # get C_l^XX from the cosmological code
        try:
            cls = deepcopy(self._current_state["source_Cl"])
        except:
            raise LoggedError(
                self.log, "No source Cl's were computed. "
                          "Are you sure that you have requested some source?")
        cls_dict = dict()
        for term, cl in cls.items():
            term_tuple = tuple(
                (lambda x: x if x == "P" else list(self.sources)[int(x) - 1])(
                    _.strip("W")) for _ in term.split("x"))
            cls_dict[term_tuple] = cl
        cls_dict["ell"] = np.arange(cls[list(cls)[0]].shape[0])
        return cls_dict

    def get_CAMBdata(self):
        """
        Get the CAMB result object (must have been requested as a requirement).

        :return: CAMB's `CAMBdata <https://camb.readthedocs.io/en/latest/results.html>`_
                 result instance for the current parameters
        """
        return self._current_state['CAMBdata']

    def get_can_provide_params(self):
        # possible derived parameters for derived_extra, excluding things that are
        # only input parameters.
        params_derived = list(get_class_methods(self.camb.CAMBparams))
        params_derived.remove("custom_source_names")
        fields = []
        for f, tp in self.camb.CAMBparams._fields_:
            if tp is ctypes.c_double and 'max_eta_k' not in f \
                    and f not in ['Alens', 'num_nu_massless']:
                fields.append(f)
        fields += ['omega_de', 'sigma8']  # only parameters from CAMBdata
        properties = get_properties(self.camb.CAMBparams)
        names = self.camb.model.derived_names + properties + fields + params_derived
        for name, mapped in self.renames.items():
            if mapped in names:
                names.append(name)
        # remove any parameters explicitly tagged as input requirements
        return set(names).difference(
            set(self._transfer_requires).union(set(self.requires)))

    def get_version(self):
        return self.camb.__version__

    def set(self, params_values_dict, state):
        # Prepare parameters to be passed: this is called from the CambTransfers instance
        args = {self.translate_param(p): v for p, v in params_values_dict.items()}
        # Generate and save
        self.log.debug("Setting parameters: %r and %r",
                       dict(args), dict(self.extra_args))
        try:
            if not self._base_params:
                base_args = args.copy()
                base_args.update(self.extra_args)
                # Remove extra args that might
                # cause an error if the associated product is not requested
                if not self.extra_attrs["WantCls"]:
                    for not_needed in getfullargspec(
                            self.camb.CAMBparams.set_for_lmax).args[1:]:
                        base_args.pop(not_needed, None)
                self._reduced_extra_args = self.extra_args.copy()
                params = self.camb.set_params(**base_args)
                # pre-set the parameters that are not varying
                for non_param_func in ['set_classes', 'set_matter_power', 'set_for_lmax']:
                    for fixed_param in getfullargspec(
                            getattr(self.camb.CAMBparams, non_param_func)).args[1:]:
                        if fixed_param in args:
                            raise LoggedError(self.log,
                                              "Trying to sample fixed theory parameter %s",
                                              fixed_param)
                        self._reduced_extra_args.pop(fixed_param, None)
                if self.extra_attrs:
                    self.log.debug("Setting attributes of CAMBparams: %r",
                                   self.extra_attrs)
                for attr, value in self.extra_attrs.items():
                    if hasattr(params, attr):
                        setattr(params, attr, value)
                    else:
                        raise LoggedError(
                            self.log,
                            "Some of the attributes to be set manually were not "
                            "recognized: %s=%s", attr, value)
                # Sources
                if getattr(self, "sources", None):
                    self.log.debug("Setting sources: %r", self.sources)
                    sources = self.camb.sources
                    source_windows = []
                    for source, window in self.sources.items():
                        function = window.pop("function", None)
                        if function == "spline":
                            source_windows.append(sources.SplinedSourceWindow(**window))
                        elif function == "gaussian":
                            source_windows.append(sources.GaussianSourceWindow(**window))
                        else:
                            raise LoggedError(self.log,
                                              "Unknown source window function type %r",
                                              function)
                        window["function"] = function
                    params.SourceWindows = source_windows
                    params.SourceTerms.limber_windows = self.limber
                self._base_params = params
            else:
                args.update(self._reduced_extra_args)
            return self.camb.set_params(self._base_params.copy(), **args)
        except self.camb.baseconfig.CAMBParamRangeError:
            if self.stop_at_error:
                raise LoggedError(self.log, "Out of bound parameters: %r",
                                  params_values_dict)
            else:
                self.log.debug("Out of bounds parameters. "
                               "Assigning 0 likelihood and going on.")
        except (self.camb.baseconfig.CAMBValueError, self.camb.baseconfig.CAMBError):
            if self.stop_at_error:
                self.log.error(
                    "Error setting parameters (see traceback below)! "
                    "Parameters sent to CAMB: %r and %r.\n"
                    "To ignore this kind of error, make 'stop_at_error: False'.",
                    dict(state["params"]), dict(self.extra_args))
                raise
        except self.camb.baseconfig.CAMBUnknownArgumentError as e:
            raise LoggedError(
                self.log,
                "Some of the parameters passed to CAMB were not recognized: %s" % str(e))
        return False

    def get_helper_theories(self):
        """
        Transfer functions are computed separately by camb.transfers, then this
        class uses the transfer functions to calculate power spectra (using A_s, n_s etc).
        """
        self._camb_transfers = CambTransfers(self, 'camb.transfers',
                                             dict(stop_at_error=self.stop_at_error),
                                             timing=self.timer)
        setattr(self._camb_transfers, _requires, self._transfer_requires)
        return {'camb.transfers': self._camb_transfers}

    def get_speed(self):
        if self._measured_speed:
            return self._measured_speed
        if not self.non_linear_sources:
            return self.speed * 10
        if {'omk', 'omegak'}.intersection(set(self._camb_transfers.input_params)):
            return self.speed / 1.5
        return self.speed * 3

    @classmethod
    def get_path(cls, path):
        return os.path.realpath(
            os.path.join(path, "code",
                         cls._camb_repo_name[cls._camb_repo_name.find("/") + 1:]))

    @classmethod
    def is_installed(cls, **kwargs):
        import platform
        if not kwargs["code"]:
            return True
        return os.path.isfile(os.path.realpath(
            os.path.join(cls.get_path(kwargs["path"]),
                         "camb", "cambdll.dll" if (
                        platform.system() == "Windows") else "camblib.so")))

    @classmethod
    def install(cls, path=None, code=True, no_progress_bars=False, **kwargs):
        log = logging.getLogger(cls.__name__)
        if not code:
            log.info("Code not requested. Nothing to do.")
            return True
        log.info("Downloading camb...")
        success = download_github_release(
            os.path.join(path, "code"), cls._camb_repo_name, cls._camb_repo_version,
            no_progress_bars=no_progress_bars, logger=log)
        if not success:
            log.error("Could not download camb.")
            return False
        camb_path = cls.get_path(path)
        log.info("Compiling camb...")
        from subprocess import Popen, PIPE
        process_make = Popen([sys.executable, "setup.py", "build_cluster"],
                             cwd=camb_path, stdout=PIPE, stderr=PIPE)
        out, err = process_make.communicate()
        if process_make.returncode:
            log.info(out.decode())
            log.info(err.decode())
            gcc_check = check_gcc_version(cls._camb_min_gcc_version, error_returns=False)
            if not gcc_check:
                cause = (" Possible cause: it looks like `gcc` does not have the correct "
                         "version number (CAMB requires %s); and `ifort` is also "
                         "probably not available." % cls._camb_min_gcc_version)
            else:
                cause = ""
            log.error("Compilation failed!" + cause)
            return False
        return True


class CambTransfers(HelperTheory):
    """
    Helper theory class that calculates transfer functions only. The result is cached
    when only initial power spectrum or non-linear model parameters change
    """

    def __init__(self, cobaya_camb, name, info, timing=None):
        self.needs_perts = False
        self.non_linear_sources = False
        super().__init__(info, name, timing=timing)
        self.cobaya_camb = cobaya_camb
        self.camb = cobaya_camb.camb
        self.speed = self.cobaya_camb.speed * 1.5

    def get_can_support_params(self):
        supported_params = self.camb.get_valid_numerical_params(
            transfer_only=True,
            dark_energy_model=self.cobaya_camb.extra_args.get('dark_energy_model'),
            recombination_model=self.cobaya_camb.extra_args.get('recombination_model')) \
                           - set(self.cobaya_camb.extra_args) \
                           - set(self.cobaya_camb.extra_attrs)

        for name, mapped in self.cobaya_camb.renames.items():
            if mapped in supported_params:
                supported_params.add(name)
        return supported_params

    def get_allow_agnostic(self):
        return False

    def must_provide(self, **requirements):
        super().must_provide(**requirements)
        opts = requirements.get('CAMB_transfers')
        if opts:
            self.non_linear_sources = opts['non_linear']
            self.needs_perts = opts['needs_perts']
        self.cobaya_camb.check_no_repeated_input_extra()

    def get_CAMB_transfers(self):
        return self._current_state['results']

    def calculate(self, state, want_derived=True, **params_values_dict):
        # Set parameters
        camb_params = self.cobaya_camb.set(params_values_dict, state)
        # Failed to set parameters but no error raised
        # (e.g. out of computationally feasible range): lik=0
        if not camb_params:
            return False
        # Compute the transfer functions
        try:
            if self.non_linear_sources:
                # only need time sources if non-linear lensing or other non-linear
                # sources. Not needed just for non-linear PK.
                results = self.camb.get_transfer_functions(camb_params,
                                                           only_time_sources=True)
            else:
                results = self.camb.get_transfer_functions(camb_params) \
                    if self.needs_perts else self.camb.get_background(camb_params)
            state['results'] = (camb_params, results)
        except self.camb.baseconfig.CAMBError as e:
            if self.stop_at_error:
                self.log.error(
                    "Computation error (see traceback below)! "
                    "Parameters sent to CAMB: %r and %r.\n"
                    "To ignore this kind of error, make 'stop_at_error: False'.",
                    dict(state["params"]), dict(self.cobaya_camb.extra_args))
                raise
            else:
                # Assumed to be a "parameter out of range" error.
                self.log.debug("Computation of cosmological products failed. "
                               "Assigning 0 likelihood and going on. "
                               "The output of the CAMB error was %s" % e)
                return False

    def initialize_with_params(self):
        if len(set(self.input_params).intersection(
                {"H0", "cosmomc_theta", "thetastar"})) > 1:
            raise LoggedError(self.log, "Can't pass more than one of H0, "
                                        "theta, cosmomc_theta to CAMB.")
        if len(set(self.input_params).intersection({"tau", "zrei"})) > 1:
            raise LoggedError(self.log, "Can't pass more than one of tau and zrei "
                                        "to CAMB.")

        super().initialize_with_params()



"""
.. module:: likelihood

:Synopsis: Likelihood class and likelihood collection
:Author: Jesus Torrado and Antony Lewis

This module defines the main :class:`Likelihood` class, from which every likelihood
usually inherits, and the :class:`LikelihoodCollection` class, which groups and manages
all the individual likelihoods.

Likelihoods inherit from :class:`~theory.Theory`, adding an additional method
to return the likelihood. As with all theories, likelihoods cache results, and the
function :meth:`LikelihoodInterface.get_current_logp` is used by :class:`model.Model` to
calculate the total likelihood. The default Likelihood implementation does the actual
calculation of the log likelihood in the `logp` function, which is then called
by :meth:`Likelihood.calculate` to save the result into the current state.

Subclasses typically just provide the `logp` function to define their likelihood result,
and use :meth:`~theory.Theory.get_requirements` to specify which inputs are needed from
other theory codes (or likelihoods). Other methods of the :class:`~theory.Theory` base
class can be used as and when needed.

"""

# Global
import sys
import traceback
import inspect
from time import sleep
from typing import Mapping, Optional, Union
from itertools import chain
import numpy as np

# Local
from cobaya.conventions import kinds, _external, _component_path, empty_dict, \
    _input_params, _output_params, _requires
from cobaya.tools import get_class, get_external_function, getfullargspec, str_to_list
from cobaya.log import LoggedError
from cobaya.component import ComponentCollection
from cobaya.theory import Theory


class LikelihoodInterface:
    """
    Interface function for likelihoods. Can descend from a :class:`~theory.Theory` class
    and this to make a likelihood (where the calculate() method stores state['logp'] for
    the current parameters), or likelihoods can directly inherit from :class:`Likelihood`
    instead.

    The get_current_logp function returns the current state's logp, and does not normally
    need to be changed.
    """

    _current_state: Mapping[str, Mapping]

    def get_current_logp(self):
        """
        Gets log likelihood for the current point

        :return:  log likelihood from the current state
        """
        return self._current_state["logp"]


class Likelihood(Theory, LikelihoodInterface):
    """Base class for likelihoods. Extends from :class:`LikelihoodInterface` and the
    general :class:`~theory.Theory` class by adding functions to return likelihoods
    functions (logp function for a given point)."""

    type: Optional[Union[list, str]] = []

    def __init__(self, info=empty_dict, name=None, timing=None, packages_path=None,
                 initialize=True, standalone=True):
        self.delay = 0
        super().__init__(info, name=name, timing=timing,
                         packages_path=packages_path, initialize=initialize,
                         standalone=standalone)
        # Make sure `types` is a list of data types, for aggregated chi2
        self.type = str_to_list(getattr(self, "type", []) or [])

    @property
    def theory(self):
        # for backwards compatibility
        return self.provider

    def logp(self, **params_values):
        """
        Computes and returns the log likelihood value.
        Takes as keyword arguments the parameter values.
        To get the derived parameters, pass a `_derived` keyword with an empty dictionary.

        Alternatively you can just implement calculate() and save the log likelihood into
        state['logp']; this may be more convenient if you also need to also calculate
        other quantities.
        """
        return None

    def marginal(self, directions=None, params_values=None):
        """
        (For analytic likelihoods only.)
        Computes the marginal likelihood.
        If nothing is specified, returns the total marginal likelihood.
        If some directions are specified (as a list, tuple or array), returns the marginal
        likelihood pdf over those directions evaluated at the given parameter values.
        """
        raise LoggedError(self.log, "Exact marginal likelihood not defined.")

    def calculate(self, state, want_derived=True, **params_values_dict):
        """
        Calculates the likelihood and any derived parameters or needs.
        Return False is the calculation fails.

        """
        derived = {} if want_derived else None
        state["logp"] = -np.inf  # in case of exception
        state["logp"] = self.logp(_derived=derived, **params_values_dict)
        self.log.debug("Computed log-likelihood = %g", state["logp"])
        if want_derived:
            state["derived"] = derived.copy()

    def wait(self):
        if self.delay:
            self.log.debug("Sleeping for %f seconds.", self.delay)
        sleep(self.delay)


class AbsorbUnusedParamsLikelihood(Likelihood):
    pass


class LikelihoodExternalFunction(Likelihood):
    def __init__(self, info, name, timing=None):
        Theory.__init__(self, info, name=name, timing=timing, standalone=False)
        # Store the external function and assign its arguments
        self.external_function = get_external_function(info[_external], name=name)
        self._self_arg = "_self"
        argspec = getfullargspec(self.external_function)
        if info.get(_input_params, []):
            setattr(self, _input_params, str_to_list(info.get(_input_params)))
        else:
            ignore_args = [self._self_arg]
            # MARKED FOR DEPRECATION IN v3.0
            ignore_args += ["_derived", "_theory"]
            # END OF DEPRECATION BLOCK
            setattr(self, _input_params,
                    [p for p in argspec.args if p not in ignore_args])
        # MARKED FOR DEPRECATION IN v3.0
        self._derived_through_arg = "_derived" in argspec.args
        # END OF DEPRECATION BLOCK
        if info.get(_output_params, []):
            setattr(self, _output_params, str_to_list(info.get(_output_params)))
        # MARKED FOR DEPRECATION IN v3.0
        elif self._derived_through_arg:
            self.log.warning(
                "The use of a `_derived` argument to deal with derived parameters will be"
                " deprecated in a future version. From now on please list your derived "
                "parameters in a list as the value of %r in the likelihood info (see "
                "documentation) and have your function return a tuple "
                "`(logp, {derived_param_1: value_1, ...})`.", _output_params)
            # BEHAVIOUR TO BE REPLACED BY ERROR:
            derived_kw_index = argspec.args[-len(argspec.defaults):].index("_derived")
            setattr(self, _output_params, argspec.defaults[derived_kw_index])
        # END OF DEPRECATION BLOCK
        else:
            setattr(self, _output_params, [])
        # Required quantities from other components
        self._uses_self_arg = self._self_arg in argspec.args
        if info.get(_requires) and not self._uses_self_arg:
            raise LoggedError(
                self.log, "If a likelihood has external requirements, declared under %r, "
                          "it needs to accept a keyword argument %r.", _requires,
                self._self_arg)
        # MARKED FOR DEPRECATION IN v3.0
        self._uses_old_theory = "_theory" in argspec.args
        if self._uses_old_theory:
            self.log.warning(
                "The use of a `_theory` argument to deal with requirements will be"
                " deprecated in a future version. From now on please indicate your "
                "requirements as the value of field %r in the likelihood info (see "
                "documentation) and have your function take a parameter `_self`.",
                _requires)
            # BEHAVIOUR TO BE REPLACED BY ERROR:
            info[_requires] = argspec.defaults[
                argspec.args[-len(argspec.defaults):].index("_theory")]
        # END OF DEPRECATION BLOCK
        self._requirements = info.get(_requires, {}) or {}
        self.log.info("Initialized external likelihood.")

    def get_requirements(self):
        return self._requirements

    def logp(self, **params_values):
        # Remove non-input params (except _derived)
        # TODO: this lines should be removed whenever input_params/reqs split is fixed
        for p in list(params_values):
            if p not in self.input_params and p != "_derived":
                params_values.pop(p)
        _derived = params_values.pop("_derived", None)
        if self._uses_self_arg:
            params_values[self._self_arg] = self
        # MARKED FOR DEPRECATION IN v3.0
        # BLOCK TO BE REMOVED
        if self._derived_through_arg:
            params_values["_derived"] = _derived
        if self._uses_old_theory:
            params_values["_theory"] = self.provider
        # END OF DEPRECATION BLOCK
        try:
            return_value = self.external_function(**params_values)
            bad_return_msg = "Expected return value `(logp, {derived_params_dict})`."
            if hasattr(return_value, "__len__"):
                logp = return_value[0]
                if self.output_params:
                    try:
                        if _derived is not None:
                            _derived.update(return_value[1])
                            params_values["_derived"] = _derived
                    except:
                        raise LoggedError(self.log, bad_return_msg)
            # MARKED FOR DEPRECATION IN v3.0 --> just after the `not` below
            elif self.output_params and not self._derived_through_arg:
                raise LoggedError(self.log, bad_return_msg)
            else:  # no return.__len__ and output_params expected
                logp = return_value
            return logp
        except Exception as ex:
            if isinstance(ex, LoggedError):
                # Assume proper error info was written before raising LoggedError
                pass
            else:
                # Print traceback
                self.log.error("".join(
                    ["-"] * 16 + ["\n\n"] +
                    list(traceback.format_exception(*sys.exc_info())) +
                    ["\n"] + ["-"] * 37))
            raise LoggedError(
                self.log, "The external likelihood '%s' failed at evaluation. "
                          "See error info on top of this message.", self.get_name())


class LikelihoodCollection(ComponentCollection):
    """
    A dictionary storing experimental likelihood :class:`Likelihood` instances index
    by their names.
    """

    def __init__(self, info_likelihood, packages_path=None, timing=None, theory=None):
        super().__init__()
        self.set_logger("likelihood")
        self.theory = theory
        # Get the individual likelihood classes
        for name, info in info_likelihood.items():
            if isinstance(name, Theory):
                name, info = name.get_name(), info
            if isinstance(info, Theory):
                self.add_instance(name, info)
            elif _external in info:
                if isinstance(info[_external], Theory):
                    self.add_instance(name, info[_external])
                elif inspect.isclass(info[_external]):
                    if not hasattr(info[_external], "get_current_logp") or \
                            not issubclass(info[_external], Theory):
                        raise LoggedError(self.log, "%s: external class likelihood must "
                                                    "be a subclass of Theory and have "
                                                    "logp, get_current_logp functions",
                                          info[_external].__name__)
                    self.add_instance(name,
                                      info[_external](info, packages_path=packages_path,
                                                      timing=timing,
                                                      standalone=False,
                                                      name=name))
                else:
                    # If it has an "external" key, wrap it up. Else, load it up
                    self.add_instance(name, LikelihoodExternalFunction(info, name,
                                                                       timing=timing))
            else:
                like_class = get_class(name, kind=kinds.likelihood,
                                       component_path=info.pop(_component_path, None))
                self.add_instance(name, like_class(info, packages_path=packages_path,
                                                   timing=timing, standalone=False,
                                                   name=name))

            if not hasattr(self[name], "get_current_logp"):
                raise LoggedError(self.log, "'Likelihood' %s is not actually a "
                                            "likelihood (no get_current_logp function)",
                                  name)

    @property
    def all_types(self):
        if not hasattr(self, "_all_types"):
            self._all_types = set(chain(
                *[str_to_list(getattr(self[like], "type", []) or []) for like in self]))
        return self._all_types
